import copy, time, traceback, logging, os, gc
import numpy as np
import pandas as pd
from sklearn.neighbors import NearestNeighbors

from ..constants import BINARY, MULTICLASS, REGRESSION, SOFTCLASS
from ..models.tabular_nn.tabular_nn_model import TabularNeuralNetModel
from ...features.feature_metadata import FeatureMetadata
from ...metrics import mean_squared_error

logger = logging.getLogger(__name__)


def format_distillation_labels(y, problem_type, num_classes=None, eps_labelsmooth=0.01):
    """ Transforms train/test label objects (y) to the correct type for distillation (smoothed regression targets for binary, one-hot labels for multiclass).
        eps_labelsmooth : truncates labels to [EPS, 1-EPS], eg. when converting binary problems -> regression
    """
    if problem_type == MULTICLASS:
        y_int = y.to_numpy()
        y = np.zeros((y_int.size, num_classes))
        y[np.arange(y_int.size),y_int] = 1
        y = pd.DataFrame(y)
    elif problem_type == BINARY:
        min_pred = 0.0
        max_pred = 1.0
        y = eps_labelsmooth + ((1-2*eps_labelsmooth)/(max_pred-min_pred)) * (y - min_pred)
    return y


def augment_data(X_train, feature_metadata: FeatureMetadata, augmentation_data=None, augment_method='spunge', augment_args=None):
    """ augment_method options: ['spunge', 'munge']
    """
    if augment_args is None:
        augment_args = {}
    if augmentation_data is not None:
        X_aug = augmentation_data
    else:
        if 'num_augmented_samples' not in augment_args:
            if 'max_size' not in augment_args:
                augment_args['max_size'] = np.inf
            augment_args['num_augmented_samples'] = int(min(augment_args['max_size'], augment_args['size_factor']*len(X_train)))

        if augment_method == 'spunge':
            X_aug = spunge_augment(X_train, feature_metadata, **augment_args)
        elif augment_method == 'munge':
            X_aug = munge_augment(X_train, feature_metadata, **augment_args)
        else:
            raise ValueError(f"unknown augment_method: {augment_method}")

    # return postprocess_augmented(X_aug, X_train)  # TODO: dropping duplicates is much more efficient, but may skew distribution for entirely-categorical data with few categories.
    logger.log(15, f"Augmented training dataset with {len(X_aug)} extra datapoints")
    return X_aug.reset_index(drop=True)


def postprocess_augmented(X_aug, X):
    """ Drops rows from augmented data that are duplicated (including duplicates that appeared in original data X). """
    X_aug = pd.concat([X, X_aug])
    X_aug.drop_duplicates(keep='first', inplace=True)
    X_aug = X_aug.tail(len(X_aug)-len(X))
    logger.log(15, f"Augmented training dataset with {len(X_aug)} extra datapoints")
    return X_aug.reset_index(drop=True, inplace=False)


def spunge_augment(X, feature_metadata: FeatureMetadata, num_augmented_samples=10000, frac_perturb=0.1, continuous_feature_noise=0.1, **kwargs):
    """ Generates synthetic datapoints for learning to mimic teacher model in distillation
        via simplified version of MUNGE strategy (that does not require near-neighbor search).

        Args:
            num_augmented_samples: number of additional augmented data points to return
            frac_perturb: fraction of features/examples that are perturbed during augmentation. Set near 0 to ensure augmented sample distribution remains closer to real data.
            continuous_feature_noise: we noise numeric features by this factor times their std-dev. Set near 0 to ensure augmented sample distribution remains closer to real data.
    """
    if frac_perturb > 1.0:
        raise ValueError("frac_perturb must be <= 1")
    logger.log(20, f"SPUNGE: Augmenting training data with {num_augmented_samples} synthetic samples for distillation...")
    num_feature_perturb = max(1, int(frac_perturb*len(X.columns)))
    X_aug = pd.concat([X.iloc[[0]].copy()]*num_augmented_samples)
    X_aug.reset_index(drop=True, inplace=True)
    continuous_types = ['float', 'int']
    continuous_featnames = feature_metadata.get_features(valid_raw_types=continuous_types)  # these features will have shuffled values with added noise

    for i in range(num_augmented_samples): # hot-deck sample some features per datapoint
        og_ind = i % len(X)
        augdata_i = X.iloc[og_ind].copy()
        num_feature_perturb_i = np.random.choice(range(1,num_feature_perturb+1))  # randomly sample number of features to perturb
        cols_toperturb = np.random.choice(list(X.columns), size=num_feature_perturb_i, replace=False)
        for feature in cols_toperturb:
            feature_data = X[feature]
            augdata_i[feature] = feature_data.sample(n=1).values[0]
        X_aug.iloc[i] = augdata_i

    for feature in X.columns:
        if feature in continuous_featnames:
            feature_data = X[feature]
            aug_data = X_aug[feature]
            noise = np.random.normal(scale=np.nanstd(feature_data)*continuous_feature_noise, size=num_augmented_samples)
            mask = np.random.binomial(n=1, p=frac_perturb, size=num_augmented_samples)
            aug_data = aug_data + noise*mask
            X_aug[feature] = pd.Series(aug_data, index=X_aug.index)

    return X_aug


def munge_augment(X, feature_metadata: FeatureMetadata, num_augmented_samples=10000, perturb_prob=0.5, s=1.0, **kwargs):
    """ Uses MUNGE algorithm to generate synthetic datapoints for learning to mimic teacher model in distillation: https://www.cs.cornell.edu/~caruana/compression.kdd06.pdf
        Args:
            num_augmented_samples: number of additional augmented data points to return
            perturb_prob: probability of perturbing each feature during augmentation. Set near 0 to ensure augmented sample distribution remains closer to real data.
            s: We noise numeric features by their std-dev divided by this factor (inverse of continuous_feature_noise). Set large to ensure augmented sample distribution remains closer to real data.
    """
    nn_dummy = TabularNeuralNetModel(path='nn_dummy', name='nn_dummy', problem_type=REGRESSION, eval_metric=mean_squared_error,
                                     hyperparameters={'num_dataloading_workers': 0, 'proc.embed_min_categories': np.inf},
                                     features = list(X.columns), feature_metadata=feature_metadata)
    processed_data = nn_dummy.process_train_data(df=nn_dummy.preprocess(X), labels=pd.Series([1]*len(X)), batch_size=nn_dummy.params['batch_size'],
                        num_dataloading_workers=0, impute_strategy=nn_dummy.params['proc.impute_strategy'],
                        max_category_levels=nn_dummy.params['proc.max_category_levels'], skew_threshold=nn_dummy.params['proc.skew_threshold'],
                        embed_min_categories=nn_dummy.params['proc.embed_min_categories'], use_ngram_features=nn_dummy.params['use_ngram_features'])
    X_vector = processed_data.dataset._data[processed_data.vectordata_index].asnumpy()
    processed_data = None
    nn_dummy = None
    gc.collect()

    neighbor_finder = NearestNeighbors(n_neighbors=2)
    neighbor_finder.fit(X_vector)
    neigh_dist, neigh_ind = neighbor_finder.kneighbors(X_vector)
    neigh_ind = neigh_ind[:,1]  # contains indices of nearest neighbors
    neigh_dist = None
    # neigh_dist = neigh_dist[:,1]  # contains distances to nearest neighbors
    neighbor_finder = None
    gc.collect()

    if perturb_prob > 1.0:
        raise ValueError("frac_perturb must be <= 1")
    logger.log(20, f"MUNGE: Augmenting training data with {num_augmented_samples} synthetic samples for distillation...")
    X = X.copy()
    X_aug = pd.concat([X.iloc[[0]].copy()]*num_augmented_samples)
    X_aug.reset_index(drop=True, inplace=True)
    continuous_types = ['float', 'int']
    continuous_featnames = feature_metadata.get_features(valid_raw_types=continuous_types)  # these features will have shuffled values with added noise
    for col in continuous_featnames:
        X_aug[col] = X_aug[col].astype(float)
        X[col] = X[col].astype(float)

    for i in range(num_augmented_samples):
        og_ind = i % len(X)
        augdata_i = X.iloc[og_ind].copy()
        neighbor_i = X.iloc[neigh_ind[og_ind]].copy()
        # dist_i = neigh_dist[og_ind]
        cols_toperturb = np.random.choice(list(X.columns), size=np.random.binomial(X.shape[1], p=perturb_prob, size=1)[0], replace=False)
        for col in cols_toperturb:
            new_val = neighbor_i[col]
            if col in continuous_featnames:
                new_val += np.random.normal(scale=np.abs(augdata_i[col]-new_val)/s)
            augdata_i[col] = new_val
        X_aug.iloc[i] = augdata_i

    return X_aug



from ....constants import BINARY, MULTICLASS, REGRESSION, SOFTCLASS

DEFAULT_ITERATIONS = 10000


def get_param_baseline(problem_type, num_classes=None):
    if problem_type == BINARY:
        return get_param_binary_baseline()
    elif problem_type in [MULTICLASS, SOFTCLASS]:
        return get_param_multiclass_baseline(num_classes=num_classes)
    elif problem_type == REGRESSION:
        return get_param_regression_baseline()
    else:
        return get_param_binary_baseline()


def get_param_binary_baseline():
    params = {
        'iterations': DEFAULT_ITERATIONS,
        'learning_rate': 0.1,
    }
    return params


def get_param_multiclass_baseline(num_classes):
    params = {
        'iterations': DEFAULT_ITERATIONS,
        'learning_rate': 0.1,
    }
    return params


def get_param_regression_baseline():
    params = {
        'iterations': DEFAULT_ITERATIONS,
        'learning_rate': 0.1,
    }
    return params



from sktmls.models.contrib import ShuffleListModel, PeriodicShuffleListModel


class TestShuffleListModel:
    def test_000_shuffle_list(self):
        slm = ShuffleListModel("test_model", "test_version", "item001", "item001_name", ["h", "e", "l", "o", "w"])

        results = slm.predict(None)
        assert "items" in results
        items = results["items"][0]

        assert items.get("id") == "item001"
        assert items.get("name") == "item001_name"
        assert "props" in items
        assert type(items.get("props").get("item001")) == list
        assert set(items.get("props").get("item001")) == {"h", "e", "l", "o", "w"}


class TestPeriodicShuffleListModel:
    def test_000_periodic_shuffle_list(self):
        slm = PeriodicShuffleListModel(
            "test_model", "test_version", "item001", "item001_name", ["h", "e", "l", "o", "w"], "daily"
        )

        results = slm.predict("1234567890")
        assert "items" in results
        items = results["items"][0]

        assert items.get("id") == "item001"
        assert items.get("name") == "item001_name"
        assert "props" in items
        assert type(items.get("props").get("item001")) == list
        assert set(items.get("props").get("item001")) == {"h", "e", "l", "o", "w"}

        second_results = slm.predict("1234567890")
        second_items = second_results["items"][0]
        assert items.get("props").get("item001") == second_items.get("props").get("item001")



import json
import os
import shlex
import subprocess
from pathlib import Path
from typing import List, TYPE_CHECKING

import joblib

from sktmls import MLSENV, MLSRuntimeENV, MLSClientError
from sktmls.config import Config

if TYPE_CHECKING:
    from sktmls.models import MLSModel, MLModelClient, ManualModel

MLS_MODEL_DIR = Path.home().joinpath("models")
MODEL_BINARY_NAME = "model.joblib"
MODEL_META_NAME = "model.json"
BUCKET = "mls-model-registry"


class ModelRegistryError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class ModelRegistry:
    """
    모델 레지스트리 클래스입니다.
    """

    def __init__(self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env

        ## Returns
        `sktmls.ModelRegistry`

        ## Example

        ```python
        model_registry = ModelRegistry(env=MLSENV.STG, runtime_env=MLSRuntimeENV.LOCAL)
        ```
        """
        if env:
            assert env in MLSENV.list_items(), "Invalid environment."
            self.__env = env
        elif os.environ.get("MLS_ENV"):
            self.__env = MLSENV[os.environ["MLS_ENV"]]
        elif os.environ.get("AWS_ENV"):
            self.__env = MLSENV[os.environ["AWS_ENV"]]
        else:
            self.__env = MLSENV.STG

        if runtime_env:
            assert runtime_env in MLSRuntimeENV.list_items(), "Invalid environment."
            self.__runtime_env = runtime_env
        elif os.environ.get("MLS_RUNTINE_ENV"):
            self.__runtime_env = MLSRuntimeENV[os.environ["MLS_RUNTINE_ENV"]]
        else:
            __HOSTNAME = os.environ.get("HOSTNAME", "").lower()
            if __HOSTNAME.startswith("bdp-dmi"):
                self.__runtime_env = MLSRuntimeENV.YE
            elif __HOSTNAME.startswith("vm-skt"):
                self.__runtime_env = MLSRuntimeENV.EDD
            else:
                self.__runtime_env = MLSRuntimeENV.LOCAL

        self.config = Config(self.__runtime_env.value)

    def save(self, mls_model: "MLSModel", force: bool = False) -> None:
        """
        모델 바이너리(model.joblib)와 정보(model.json)를 MLS 모델 레지스트리에 등록합니다.

        `sktmls.models.MLSModel.save`와 동일하게 동작합니다.

        ## Args

        - mls_model: (`sktmls.models.MLSModel`) 모델 객체
        - force: (bool) 이미 모델 레지스트리에 등록된 경우 덮어 쓸 것인지 여부 (기본값: `False`)

        ## Example

        ```python
        model_registry = ModelRegistry(env=MLSENV.STG)
        model_registry.save(gbm_model)
        ```
        """
        model_path = MLS_MODEL_DIR.joinpath(mls_model.model_name, mls_model.model_version)
        model_binary_path = model_path.joinpath(MODEL_BINARY_NAME)
        model_meta_path = model_path.joinpath(MODEL_META_NAME)

        if self.config.MLS_RUNTIME_ENV != "LOCAL" and mls_model.model_lib == "autogluon":
            mls_model.set_mms_path()

        model_path.mkdir(parents=True, exist_ok=True)
        joblib.dump(mls_model, model_binary_path)
        with model_meta_path.open("w") as f:
            json.dump(
                {
                    "name": mls_model.model_name,
                    "version": mls_model.model_version,
                    "model_lib": mls_model.model_lib,
                    "model_lib_version": mls_model.model_lib_version,
                    "model_data": f"/models/{mls_model.model_name}/{mls_model.model_version}/{MODEL_BINARY_NAME}",
                    "features": mls_model.features,
                    "class": mls_model.__class__.__name__,
                    "metric": getattr(mls_model, "metric", None),
                    "performance": getattr(mls_model, "performance", None),
                    "feature_importance": getattr(mls_model, "feature_importance", None),
                    "model_info": getattr(mls_model, "model_info", None),
                },
                f,
            )

        if self.config.MLS_RUNTIME_ENV == "LOCAL":
            return

        s3_path = BUCKET
        if self.__env in (MLSENV.STG, MLSENV.PRD):
            s3_path = f"{BUCKET}-{self.__env.value}"
        s3_path = f"{s3_path}/{mls_model.model_name}"

        force_option = "-f" if force else ""

        process_mkdir = subprocess.Popen(
            shlex.split(f"hdfs dfs {self.config.HDFS_OPTIONS} -mkdir -p s3a://{s3_path}"),
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        process_mkdir.wait()
        if process_mkdir.returncode != 0:
            raise ModelRegistryError(f"{mls_model.model_name}/{mls_model.model_version}: S3 폴더 생성에 실패했습니다.")

        process_put = subprocess.Popen(
            shlex.split(f"hdfs dfs {self.config.HDFS_OPTIONS} -put {force_option} {model_path} s3a://{s3_path}"),
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        process_put.wait()
        if process_put.returncode != 0:
            raise ModelRegistryError(f"{mls_model.model_name}/{mls_model.model_version}: S3 복사가 실패했습니다.")

        if mls_model.model_lib == "autogluon":
            mls_model.set_local_path()

    def save_and_deploy(
        self, mls_model: "MLSModel", ml_model_client: "MLModelClient", force: bool = False
    ) -> "ManualModel":
        """
        모델 바이너리(model.joblib)와 정보(model.json)를 MLS 모델 레지스트리에 등록하고, 해당 바이너리를 참조하는 MLS의 일반 모델(ManualModel)을 생성(deploy)합니다.

        일반 모델은 deploy 후 버킷에 연결되어 정상 동작하기 까지 약 한 시간이 소요됩니다.

        ## Args

        - mls_model: (`sktmls.models.MLSModel`) 모델 객체
        - ml_model_client: (`sktmls.models.MLModelClient`) 모델 클라이언트
        - force: (bool) 이미 모델 레지스트리에 등록된 경우 덮어 쓸 것인지 여부 (기본값: `False`)

        ## Example

        ```python
        ml_model_client = MLModelClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        model_registry = ModelRegistry(env=MLSENV.STG)
        model_registry.save_and_deploy(gbm_model, ml_model_client)
        ```
        """
        self.save(mls_model, force)

        try:
            return ml_model_client.create_manual_model(
                name=mls_model.model_name,
                version=mls_model.model_version,
                model_lib=mls_model.model_lib,
                model_data=f"/models/{mls_model.model_name}/{mls_model.model_version}/model.joblib",
                features=",".join(mls_model.features),
                model_meta={
                    "name": mls_model.model_name,
                    "version": mls_model.model_version,
                    "model_lib": mls_model.model_lib,
                    "model_lib_version": mls_model.model_lib_version,
                    "model_data": f"/models/{mls_model.model_name}/{mls_model.model_version}/model.joblib",
                    "features": mls_model.features,
                },
            )
        except MLSClientError as e:
            if not force:
                raise e

            manual_model = ml_model_client.get_manual_model(name=mls_model.model_name, version=mls_model.model_version)
            return ml_model_client.update_manual_model(
                manual_model=manual_model,
                model_lib=mls_model.model_lib,
                model_data=f"/models/{mls_model.model_name}/{mls_model.model_version}/model.joblib",
                features=",".join(mls_model.features),
                model_meta={
                    "name": mls_model.model_name,
                    "version": mls_model.model_version,
                    "model_lib": mls_model.model_lib,
                    "model_lib_version": mls_model.model_lib_version,
                    "model_data": f"/models/{mls_model.model_name}/{mls_model.model_version}/model.joblib",
                    "features": mls_model.features,
                },
            )

    def list_models(self) -> List[str]:
        """
        MLS 모델 레지스트리에 등록된 모든 모델 이름을 리스트로 가져옵니다.

        ## Returns
        list(str)

        ## Example

        ```python
        model_registry = ModelRegistry(env=MLSENV.STG)
        model_names = model_registry.list_models()
        ```
        """
        if self.config.MLS_RUNTIME_ENV == "LOCAL":
            return [x.name for x in MLS_MODEL_DIR.iterdir() if x.is_dir()]

        s3_path = BUCKET
        if self.__env in (MLSENV.STG, MLSENV.PRD):
            s3_path = f"{BUCKET}-{self.__env.value}"

        s3_path = f"s3a://{s3_path}/"
        process = subprocess.Popen(
            shlex.split(f"hdfs dfs {self.config.HDFS_OPTIONS} -ls {s3_path}"),
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        process.wait()
        if process.returncode != 0:
            raise ModelRegistryError(f"Listing models in ({s3_path}) is FAILED.")

        output = [row.split(s3_path)[-1] for row in process.stdout.read().decode().split("\n") if s3_path in row]
        return output

    def list_versions(self, model_name: str) -> List[str]:
        """
        MLS 모델 레지스트리에 등록된 모델의 모든 버전을 리스트로 가져옵니다.

        ## Args

        - model_name: (str) 모델 이름

        ## Returns
        list(str)

        ## Example

        ```python
        model_registry = ModelRegistry(env=MLSENV.STG)
        model_versions = model_registry.list_versions("hello_model")
        ```
        """
        if self.config.MLS_RUNTIME_ENV == "LOCAL":
            return [x.name for x in MLS_MODEL_DIR.joinpath(model_name).iterdir() if x.is_dir()]

        s3_path = BUCKET
        if self.__env in (MLSENV.STG, MLSENV.PRD):
            s3_path = f"{BUCKET}-{self.__env.value}"
        s3_path = f"s3a://{s3_path}/{model_name}/"

        process = subprocess.Popen(
            shlex.split(f"hdfs dfs {self.config.HDFS_OPTIONS} -ls {s3_path}"),
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        process.wait()
        if process.returncode != 0:
            raise ModelRegistryError(f"Listing versions in ({s3_path}) is FAILED.")

        output = [row.split(s3_path)[-1] for row in process.stdout.read().decode().split("\n") if s3_path in row]
        return output

    def load(self, model_name: str, model_version: str) -> "MLSModel":
        """
        MLS 모델 레지스트리로부터 모델 객체를 가져옵니다.

        ## Args
        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전

        ## Returns
        `sktmls.models.MLSModel`

        ## Example

        ```python
        model_registry = ModelRegistry(env=MLSENV.STG)
        hello_model_v1 = model_registry.load(model_name="hello_model", model_version="v1")
        results = hello_model_v1.predict([1, 2, 3])
        ```
        """
        model_path = MLS_MODEL_DIR.joinpath(model_name, model_version)
        model_name_path = MLS_MODEL_DIR.joinpath(model_name)
        model_binary_path = model_path.joinpath(MODEL_BINARY_NAME)

        if self.config.MLS_RUNTIME_ENV == "LOCAL":
            mls_model = joblib.load(model_binary_path)
            if mls_model.model_lib == "autogluon":
                mls_model.set_local_path()
            return mls_model

        s3_path = BUCKET
        if self.__env in (MLSENV.STG, MLSENV.PRD):
            s3_path = f"{BUCKET}-{self.__env.value}"
        s3_path = f"s3a://{s3_path}/{model_name}/{model_version}"

        model_name_path.mkdir(parents=True, exist_ok=True)

        process = subprocess.Popen(
            shlex.split(f"hdfs dfs {self.config.HDFS_OPTIONS} -get -f {s3_path} {model_name_path}"),
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        process.wait()
        if process.returncode != 0:
            raise ModelRegistryError(f"{model_name}/{model_version}: 모델 가져오기가 실패했습니다.")

        mls_model = joblib.load(model_binary_path)
        if mls_model.model_lib == "autogluon":
            mls_model.set_local_path()

        return mls_model



from typing import List, Union
from enum import Enum

from dateutil import parser

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSClientError, MLSResponse
from sktmls.datasets import Dataset
from sktmls.components import Component

MLS_MODELS_API_URL = "/api/v1/models"
MLS_DYNAMODB_TABLES_API_URL = "/api/v1/dynamodb/tables"
MLS_MODEL_REGISTRY_ARTIFACTS_API_URL = "/api/v1/model_registry/artifacts/info"

MODEL_LIBS = ["sklearn", "xgboost", "lightgbm", "catboost", "rule", "etc"]


class MLModel:
    """
    ML모델 클래스 입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - version: (str) ML모델 버전
            - creator: (str) ML모델 생성 계정명
            - description: (str) ML모델 설명
            - model_type: (str) ML모델 타입 (`automl` | `manual`)
            - table: (str) ML모델 테이블
            - model_data: (str) ML모델 데이터
            - status: (`sktmls.models.MLModelStatus`) ML모델 상태
            - created_at: (datetime) 생성일시
            - updated_at: (datetime) 수정일시

        ## Returns
        `sktmls.models.MLModel`
        """
        self.id = kwargs.get("id")
        self.name = kwargs.get("name")
        self.version = kwargs.get("version")
        self.description = kwargs.get("description")
        self.creator = kwargs.get("creator")
        self.model_type = kwargs.get("model_type")
        self.table = kwargs.get("table")
        self.model_data = kwargs.get("model_data")
        self.status = kwargs.get("status")
        try:
            self.created_at = parser.parse(kwargs.get("created_at"))
        except TypeError:
            self.created_at = None
        try:
            self.updated_at = parser.parse(kwargs.get("updated_at"))
        except TypeError:
            self.updated_at = None

    def __str__(self) -> str:
        return self.name

    def get(self) -> dict:
        return self.__dict__

    def reset(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class AutoMLModel(MLModel):
    """
    AutoML모델 클래스 입니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - model_type: (str) ML모델 타입 (`automl` | `manual`)
            - version: (str) ML모델 버전
            - description: (str) ML모델 설명
            - creator: (str) ML모델 생성 계정명
            - status: (`sktmls.models.MLModelStatus`) ML모델 상태
            - model_meta: (dict) ML모델 메타정보
            - dataset_id: (int) ML 데이터셋 고유 ID
            - dataset_name: (str) ML 데이터셋 이름
            - automl_model_info: (dict) AutoML모델 정보

        ## Returns
        `sktmls.models.AutoMLModel`
        """

        super().__init__(**kwargs)
        self.dataset_id = kwargs.get("dataset_id")
        self.dataset_name = kwargs.get("dataset_name")
        self.automl_model_info = kwargs.get("automl_model_info")

    def get_error(self):
        """
        모델 학습 시 에러 발생 원인을 조회한다. (모델 학습 실패 시에만 값 존재)

        ## Returns
        str
        """
        return (self.automl_model_info or {}).get("error")


class ManualModel(MLModel):
    """
    일반 모델 클래스이자, 모델 배포 객체입니다. 위 객체를 생성해 배치 추론 테이블 또는 모델 레지스트리에 등록된 모델 아티팩트를 서비스 가능하도록 배포합니다.
    배포된 모델만 버킷에 연결하여 서비스 가능합니다.
    """

    def __init__(self, **kwargs):
        """
        ## Args

        - kwargs
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - model_type: (str) ML모델 타입 (`automl` | `manual`)
            - version: (str) ML모델 버전
            - description: (str) ML모델 설명
            - creator: (str) ML모델 생성 계정명
            - status: (`sktmls.models.MLModelStatus`) ML모델 상태
            - model_meta: (dict) ML모델 메타정보
            - components: list(`sktmls.components.Component`) ML모델 컴포넌트 리스트
            - table: (str) ML모델 테이블
            - model_lib: (str) ML모델 라이브러리
            - model_data: (str) ML모델 데이터
            - features: (str) ML모델에 사용된 피쳐
            - is_enabled: (bool) 활성상태 여부

        ## Returns
        `sktmls.models.ManualModel`
        """

        super().__init__(**kwargs)
        self.model_meta = kwargs.get("model_meta")
        self.components = (
            [Component(**component) for component in kwargs.get("components")]
            if kwargs.get("components")
            else kwargs.get("components")
        )
        self.model_lib = kwargs.get("model_lib")
        self.features = kwargs.get("features")
        self.is_enabled = kwargs.get("is_enabled")


class MLModelStatus(Enum):
    IN_USE = "IN_USE"
    NOT_IN_USE = "NOT_IN_USE"
    AUTOML_TRAINING = "TRAINING"
    AUTOML_DONE = "DONE"
    AUTOML_FAILED = "FAILED"
    UNKNOWN = "UNKNOWN"


class MLModelClient(MLSClient):
    def __init__(
        self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, username: str = None, password: str = None
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env
        - $MLS_USERNAME: username
        - $MLS_PASSWORD: password

        ## Returns
        `sktmls.models.MLModelClient`

        ## Example

        ```python
        ml_model_client = MLModelClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def create_automl_model(self, dataset: Dataset, name: str, version: str, description: str) -> AutoMLModel:
        """
        AutoML 모델을 생성합니다.

        ## Args

        - dataset: (sktmls.datasets.Dataset) ML 데이터셋
        - name: (str) 생성할 AutoML 모델 이름
        - version: (str) 생성할 AutoML 모델 버전
        - description: (str) 생성할 AutoML 설명

        ## Returns
        `sktmls.models.AutoMLModel`

        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - creator: (str) ML모델 생성 계정명
        - description: (str) ML모델 설명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - dataset_id: (int) ML 데이터셋 고유 ID
        - dataset_name: (str) ML 데이터셋 이름
        - automl_model_info: (dict) AutoML모델 정보

        ## Example

        ```python
        automl_model = ml_model_client.create_automl_model(
            dataset=dataset,
            name="automl_test_model",
            version="v1",
            description="test_model"
        )
        ```
        """
        assert type(dataset) == Dataset

        data = {
            "model_type": "automl",
            "name": name,
            "version": version,
            "description": description,
            "dataset_id": dataset.id,
        }
        response = self._request(method="POST", url=MLS_MODELS_API_URL, data=data).results
        return self.get_automl_model(id=response.get("id"))

    def create_manual_model(
        self,
        name: str,
        version: str,
        description: str = None,
        creator: str = None,
        model_lib: str = None,
        model_data: str = None,
        model_meta: dict = None,
        features: str = None,
        table: str = None,
        components: List[Component] = None,
    ) -> ManualModel:
        """
        일반 모델을 생성합니다. 모델을 버킷에 연결 가능하도록 deploy하는 작업입니다.

        일반 모델은 deploy 후 버킷에 연결되어 정상 동작하기 까지 약 한 시간이 소요됩니다.

        ## Args

        - name: (str) 생성할 일반 모델 이름
        - version: (str) 생성할 일반 모델 버전
        - description: (optional) (str) 생성할 일반 모델 설명
        - creator: (optional) (str) 일반 모델 생성 계정명
        - model_lib: (optional) (str) 일반 모델 라이브러리
        - model_data: (optional) (str) 일반 모델 데이터
        - model_meta: (optional) (dict) 모델 레지스트리에 등록된 메타정보(model.json)
        - features: (optional) (str) 일반 모델에 사용된 피쳐
        - table: (optional) (str) 일반 모델 테이블
        - components: (optional) list(`sktmls.components.Component`) 일반 모델 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        component_1 = component_client.get_component(name="test_component_1")
        component_2 = component_client.get_component(name="test_component_2")
        manual_model = ml_model_client.create_manual_model(
            name="manual_test_model",
            version="v1",
            description="test_model",
            creator="mls_user",
            model_lib="lightgbm",
            model_data="/models/manual_test_model/v1/model.joblib",
            model_meta={
                "name": "manual_test_model",
                "version": "v1",
                "model_lib": "lightgbm",
                "model_lib_version": "2.3.1",
                "model_data": "/models/manual_test_model/v1/model.joblib",
                "features": ["age", "app_use_traffic_youtube", "svc_scrb_period"],
            },
            features="age,app_use_traffic_youtube,svc_scrb_period",
            table="test_model_table",
            components=[component_1, component_2]
        )
        ```
        """

        data = {
            "model_type": "manual",
            "name": name,
            "version": version,
        }
        if description:
            data["description"] = description
        if creator:
            data["creator"] = creator
        if model_lib:
            assert model_lib in MODEL_LIBS, "허용되지 않는 `model_lib`값입니다."
            data["model_lib"] = model_lib
        if model_data:
            data["model_data"] = model_data
        if model_meta:
            data["model_meta"] = model_meta
        if features:
            data["features"] = features
        if table:
            data["table"] = table
        if components:
            for component in components:
                assert type(component) == Component, "`components`는 list(`sktmls.components.Component`) 형태여야 합니다."
            data["components"] = [component.name for component in components]

        response = self._request(method="POST", url=MLS_MODELS_API_URL, data=data).results
        return self.get_manual_model(id=response.get("id"))

    def list_automl_models(self, **kwargs) -> List[AutoMLModel]:
        """
        AutoML모델 리스트를 가져옵니다.

        ## Args

        - kwargs: (optional) (dict) 쿼리 조건
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - version: (str) ML모델 이름
            - query: (str) 검색 문자
            - page: (int) 페이지 번호

        ## Returns
        list(`sktmls.models.AutoMLModel`)

        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - creator: (str) ML모델 생성 계정명
        - description: (str) ML모델 설명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - dataset_id: (int) ML 데이터셋 고유 ID
        - dataset_name: (str) ML 데이터셋 이름
        - automl_model_info: (dict) AutoML모델 정보

        ## Example

        ```python
        automl_models = ml_model_client.list_automl_models()
        ```
        """

        kwargs.update({"model_type": "automl"})
        response = self._request(method="GET", url=MLS_MODELS_API_URL, params=kwargs)
        return [AutoMLModel(**ml_model) for ml_model in response.results]

    def list_manual_models(self, **kwargs) -> List[ManualModel]:
        """
        일반 모델 리스트를 가져옵니다.

        ## Args

        - kwargs: (optional) (dict) 쿼리 조건
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - version: (str) ML모델 이름
            - query: (str) 검색 문자
            - page: (int) 페이지 번호

        ## Returns
        list(`sktmls.models.ManualModel`)

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        manual_models = ml_model_client.list_manual_models()
        ```
        """
        kwargs.update({"model_type": "manual"})
        response = self._request(method="GET", url=MLS_MODELS_API_URL, params=kwargs)
        return [ManualModel(**ml_model) for ml_model in response.results]

    def get_automl_model(self, id: int = None, name: str = None, version: str = None) -> AutoMLModel:
        """
        AutoML모델 정보를 가져옵니다.

        ## Args: `id` 또는 (`name` and `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - version: (str) ML모델 이름

        ## Returns
        `sktmls.models.AutoMLModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - dataset_id: (int) ML 데이터셋 고유 ID
        - dataset_name: (str) ML 데이터셋 이름
        - automl_model_info: (dict) AutoML모델 정보

        ## Example

        ```python
        automl_model_by_id = ml_model_client.get_automl_model(id=3)
        automl_model_by_name_and_version = ml_model_client.get_automl_model(name="my_automl_model", version="my_automl_version")
        ```
        """
        assert id or (name and version), "`id` 또는 (`name`, `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        automl_models = self.list_automl_models(id=id, name=name, version=version)

        if len(automl_models) == 0:
            raise MLSClientError(code=404, msg="ML모델이 없습니다.")

        response = self._request(method="GET", url=f"{MLS_MODELS_API_URL}/{automl_models[0].id}")
        return AutoMLModel(**response.results)

    def get_manual_model(self, id: int = None, name: str = None, version: str = None) -> ManualModel:
        """
        일반 모델 정보를 가져옵니다.

        ## Args: `id` 또는 (`name` and `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - version: (str) ML모델 이름

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        manual_model_by_id = ml_model_client.get_manual_model(id=3)
        manual_model_by_name_and_version = ml_model_client.get_manual_model(name="my_manual_model", version="my_manual_version")
        ```
        """
        assert id or (name and version), "`id` 또는 (`name`, `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다."

        manual_models = self.list_manual_models(id=id, name=name, version=version)
        if len(manual_models) == 0:
            raise MLSClientError(code=404, msg="ML모델이 없습니다.")

        return ManualModel(**self._request(method="GET", url=f"{MLS_MODELS_API_URL}/{manual_models[0].id}").results)

    def update_manual_model(
        self,
        manual_model: ManualModel,
        name: str = None,
        version: str = None,
        description: str = None,
        creator: str = None,
        model_lib: str = None,
        model_data: str = None,
        model_meta: dict = None,
        features: str = None,
        table: str = None,
        components: List[Component] = None,
    ) -> ManualModel:
        """
        일반 모델 정보를 수정합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체
        - name: (str) (optional) 수정할 일반 모델 이름
        - version: (str) (optional) 수정할 일반 모델 버전
        - description: (optional) (str) 수정할 일반 모델 설명
        - creator: (optional) (str) 수정할 일반 모델 계정명
        - model_lib: (optional) (str) 수정할 일반 모델 라이브러리
        - model_data: (optional) (str) 수정할 일반 모델 데이터
        - model_meta: (optional) (dict) 수정할 모델 레지스트리에 등록된 메타정보(model.json)
        - features: (optional) (str) 수정할 일반 모델에 사용된 피쳐
        - table: (optional) (str) 수정할 일반 모델 테이블
        - components: (optional) list(`sktmls.components.Component`) 수정할 일반 모델 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        manual_model = ml_model_client.get_manual_model(name="my_manual_model", version="v1")
        manual_model = ml_model_client.update_manual_model(
            manual_model=manual_model,
            name="my_new_manual_model",
            description="updated manual model",
            model_data="/models/my_new_manual_model/v1/model.joblib",
        )
        ```
        """
        assert type(manual_model) == ManualModel

        data = {
            "name": manual_model.name,
            "version": manual_model.version,
            "model_type": "manual",
        }
        if manual_model.description is not None:
            data["description"] = manual_model.description
        if manual_model.creator is not None:
            data["creator"] = manual_model.creator
        if manual_model.model_lib is not None:
            data["model_lib"] = manual_model.model_lib
        if manual_model.model_data is not None:
            data["model_data"] = manual_model.model_data
        if manual_model.model_meta is not None:
            data["model_meta"] = manual_model.model_meta
        if manual_model.features is not None:
            data["features"] = manual_model.features
        if manual_model.table is not None:
            data["table"] = manual_model.table
        if manual_model.components is not None:
            data["components"] = [component.get("name") for component in manual_model.components]

        if name:
            data["name"] = name
        if version:
            data["version"] = version
        if description is not None:
            data["description"] = description
        if creator is not None:
            data["creator"] = creator
        if model_lib is not None:
            assert model_lib in MODEL_LIBS, "허용되지 않는 `model_lib`값입니다."
            data["model_lib"] = model_lib
        if model_data is not None:
            data["model_data"] = model_data
        if model_meta is not None:
            data["model_meta"] = model_meta
        if features is not None:
            data["features"] = features
        if table is not None:
            data["table"] = table
        if components is not None:
            for component in components:
                assert type(component) == Component, "`components`는 list(`sktmls.components.Component`) 형태여야 합니다."
            data["components"] = [component.name for component in components]

        updated_model_id = self._request(
            method="PUT", url=f"{MLS_MODELS_API_URL}/{manual_model.id}", data=data
        ).results.get("id")

        manual_model.reset(**self.get_manual_model(id=updated_model_id).get())
        return manual_model

    def delete_model(self, model: Union[AutoMLModel, ManualModel], delete_table: False) -> MLSResponse:
        """
        모델 삭제합니다.

        ## Args

        - model: (`sktmls.models.AutoMLModel` or `sktmls.models.ManualModel`) AutoML 혹은 일반 모델 객체
        - delete_table: (optional) (bool) manual_model의 경우 모델의 연결된 dynamodb table을 함께 지우는 옵션. (기본값: `False`)

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```python
        ml_model_client.delete_model(model=model, delete_table=True)
        ```
        """

        assert type(model) in [
            AutoMLModel,
            ManualModel,
        ], "`model`은 `sktmls.models.AutoMLModel` 혹은 `sktmls.models.ManualModel` 타입이어야 합니다."

        response = self._request(method="DELETE", url=f"{MLS_MODELS_API_URL}/{model.id}")

        if delete_table and model.table:
            assert type(model) == ManualModel, "ManualModel의 경우에만 테이블 삭제 옵션을 사용할 수 있습니다"
            self._request(method="DELETE", url=f"{MLS_DYNAMODB_TABLES_API_URL}/{model.table}")

        return response

    def update_manual_model_meta(
        self,
        manual_model: ManualModel,
        model_meta: dict,
        partial_update: bool = False,
    ) -> ManualModel:
        """
        일반 모델의 모델 메타 정보를 수정합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체
        - model_meta: (dict) 수정할 모델 메타 정보
        - partial_update: (optional) (dict) 기존 모델 메타에 덮어씌울지(True) 기존 모델 메타를 교체할지(False) 여부 (기본값: `False`)

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        manual_model = ml_model_client.get_manual_model(name="my_manual_model", version="v1")
        manual_model = ml_model_client.update_manual_model_meta(
            manual_model=manual_model,
            model_meta={
                "model_lib": "xgboost",
                "model_lib_version": "1.2.0",
            },
            partial_update = True
        )
        manual_model_without_meta = ml_model_client.update_manual_model_meta(
            manual_model=manual_model,
            model_meta={}
        )

        ```
        """
        assert type(manual_model) == ManualModel

        if not partial_update:
            return self.update_manual_model(manual_model=manual_model, model_meta=model_meta)

        data = manual_model.model_meta
        data.update(model_meta)

        return self.update_manual_model(manual_model=manual_model, model_meta=data)

    def update_manual_model_components(
        self,
        manual_model: ManualModel,
        components: List[Component],
    ) -> ManualModel:
        """
        일반 모델의 컴포넌트 정보를 수정합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체
        - components: list(`sktmls.components.Component`) 수정할 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        component = component_client.get_component(name="test_component")
        manual_model = ml_model_client.get_manual_model(name="manual_test_model", version="v1")

        updated_manual_model = ml_model_client.update_manual_model_components(
            manual_model=manual_model,
            components=[component],
        )
        no_component_manual_model = ml_model_client.update_manual_model_components(
            manual_model=manual_model,
            components=[],
        )

        ```
        """
        assert type(manual_model) == ManualModel
        for component in components:
            assert type(component) == Component, "`components`는 list(`sktmls.components.Component`) 형태여야 합니다."

        return self.update_manual_model(manual_model=manual_model, components=components)

    def is_batch_prediction_ready(self, manual_model: ManualModel) -> bool:
        """
        일반 모델에 등록된 배치추론 테이블이 존재하는지와 업로드가 끝났는지 체크합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체

        ## Returns
        `bool`

        ## Example

        ```python
        my_model = ml_model_client.get_manual_model(name="manual_test_model", version="v1")
        ml_model_client.is_batch_prediction_ready(manual_model=my_model)
        ```
        """
        assert type(manual_model) == ManualModel

        table = manual_model.table
        if not table:
            raise MLSClientError(code=404, msg="해당 일반모델에 배치추론 테이블이 연결되지 않았습니다.")

        if not self._request(method="GET", url=f"{MLS_DYNAMODB_TABLES_API_URL}/{table}/is_active").results["is_active"]:
            return False

        return not self._request(method="GET", url=f"{MLS_DYNAMODB_TABLES_API_URL}/{table}/is_loading").results[
            "is_loading"
        ]

    def is_online_prediction_ready(self, manual_model: ManualModel) -> bool:
        """
        일반 모델에 등록된 아티팩트(model.jobilb)이 존재하는지 확인합니다.

        ## Args

        - model: (`sktmls.models.ManualModel`) 일반 모델 객체

        ## Returns
        `bool`

        ## Example

        ```python
        my_model = ml_model_client.get_manual_model(name="manual_test_model", version="v1")
        ml_model_client.is_online_prediction_ready(manual_model=my_model)
        ```
        """
        assert type(manual_model) == ManualModel

        data = {"model_name": manual_model.name, "model_version": manual_model.version}
        artifact = self._request(method="GET", url=MLS_MODEL_REGISTRY_ARTIFACTS_API_URL, params=data).results.get(
            "info"
        )

        if not artifact:
            return False
        return True

    def create_model_deployment(
        self,
        name: str,
        version: str,
        description: str = None,
        creator: str = None,
        model_lib: str = None,
        model_data: str = None,
        model_meta: dict = None,
        features: str = None,
        table: str = None,
        components: List[Component] = None,
    ) -> ManualModel:
        """
        일반 모델을 배포(배포 객체를 생성)합니다.

        위 메소드는 create_manual_model과 같은 기능을 수행합니다.

        모델을 버킷에 연결 가능하도록 deploy하는 작업입니다.
        일반 모델은 deploy 후 버킷에 연결되어 정상 동작하기 까지 약 한 시간이 소요됩니다.

        ## Args

        - name: (str) 생성할 일반 모델 이름
        - version: (str) 생성할 일반 모델 버전
        - description: (optional) (str) 생성할 일반 모델 설명
        - creator: (optional) (str) 일반 모델 생성 계정명
        - model_lib: (optional) (str) 일반 모델 라이브러리
        - model_data: (optional) (str) 일반 모델 데이터
        - model_meta: (optional) (dict) 모델 레지스트리에 등록된 메타정보(model.json)
        - features: (optional) (str) 일반 모델에 사용된 피쳐
        - table: (optional) (str) 일반 모델 테이블
        - components: (optional) list(`sktmls.components.Component`) 일반 모델 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        component_1 = component_client.get_component(name="test_component_1")
        component_2 = component_client.get_component(name="test_component_2")
        model_deployment = ml_model_client.create_model_deployment(
            name="manual_test_model",
            version="v1",
            description="test_model",
            creator="mls_user",
            model_lib="lightgbm",
            model_data="/models/manual_test_model/v1/model.joblib",
            model_meta={
                "name": "manual_test_model",
                "version": "v1",
                "model_lib": "lightgbm",
                "model_lib_version": "2.3.1",
                "model_data": "/models/manual_test_model/v1/model.joblib",
                "features": ["age", "app_use_traffic_youtube", "svc_scrb_period"],
            },
            features="age,app_use_traffic_youtube,svc_scrb_period",
            table="test_model_table",
            components=[component_1, component_2]
        )
        ```
        """

        return self.create_manual_model(
            name, version, description, creator, model_lib, model_data, model_meta, features, table, components
        )

    def get_model_deployment(self, id: int = None, name: str = None, version: str = None) -> ManualModel:
        """
        일반 모델 배포를 가져옵니다.
        위 메소드는 get_manual_model과 같은 기능을 수행합니다.

        ## Args: `id` 또는 (`name` and `version`) 중 한 개 이상의 값이 반드시 전달되어야 합니다.

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - version: (str) ML모델 이름

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        model_deployment_by_id = ml_model_client.get_model_deployment(id=3)
        model_deployment_by_name_and_version = ml_model_client.get_model_deployment(name="my_manual_model", version="my_manual_version")
        ```
        """
        return self.get_manual_model(id, name, version)

    def update_model_deployment(
        self,
        model_deployment: ManualModel,
        name: str = None,
        version: str = None,
        description: str = None,
        creator: str = None,
        model_lib: str = None,
        model_data: str = None,
        model_meta: dict = None,
        features: str = None,
        table: str = None,
        components: List[Component] = None,
    ) -> ManualModel:
        """
        일반 모델 배포를 수정합니다.
        위 메소드는 update_manual_model과 같은 기능을 수행합니다.

        ## Args

        - model_deployment: (`sktmls.models.ManualModel`) 일반 모델 객체
        - name: (str) (optional) 수정할 일반 모델 이름
        - version: (str) (optional) 수정할 일반 모델 버전
        - description: (optional) (str) 수정할 일반 모델 설명
        - creator: (optional) (str) 수정할 일반 모델 계정명
        - model_lib: (optional) (str) 수정할 일반 모델 라이브러리
        - model_data: (optional) (str) 수정할 일반 모델 데이터
        - model_meta: (optional) (dict) 수정할 모델 레지스트리에 등록된 메타정보(model.json)
        - features: (optional) (str) 수정할 일반 모델에 사용된 피쳐
        - table: (optional) (str) 수정할 일반 모델 테이블
        - components: (optional) list(`sktmls.components.Component`) 수정할 일반 모델 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        model_deployment = ml_model_client.get_model_deployment(name="my_manual_model", version="v1")
        model_deployment = ml_model_client.update_model_deployment(
            model_deployment=model_deployment,
            name="my_new_manual_model_deployment",
            description="updated manual model deployment",
            model_data="/models/my_new_manual_model/v1/model.joblib",
        )
        ```
        """
        return self.update_manual_model(
            model_deployment,
            name,
            version,
            description,
            creator,
            model_lib,
            model_data,
            model_meta,
            features,
            table,
            components,
        )

    def delete_model_deployment(self, model_deployment: ManualModel, delete_table: False) -> MLSResponse:
        """
        모델 배포를 삭제합니다.
        위 메소드는 delete_model과 같은 기능을 수행합니다.

        ## Args

        - model_deployment: (`sktmls.models.ManualModel`) 일반 모델 객체
        - delete_table: (optional) (bool) manual_model의 경우 모델의 연결된 dynamodb table을 함께 지우는 옵션. (기본값: `False`)

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```python
        ml_model_client.delete_model_deployment(model_deployment=model_deployment, delete_table=True)
        ```
        """

        return self.delete_model(model_deployment, delete_table)

    def update_model_deployment_meta(
        self,
        model_deployment: ManualModel,
        model_meta: dict,
        partial_update: bool = False,
    ) -> ManualModel:
        """
        일반 모델 배포의 모델 메타 정보를 수정합니다.
        위 메소드는 update_manual_model_meta와 같은 기능을 수행합니다.

        ## Args

        - model_deployment: (`sktmls.models.ManualModel`) 일반 모델 객체
        - model_meta: (dict) 수정할 모델 메타 정보
        - partial_update: (optional) (dict) 기존 모델 메타에 덮어씌울지(True) 기존 모델 메타를 교체할지(False) 여부 (기본값: `False`)

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        model_deployment = ml_model_client.get_model_deployment(name="my_manual_model", version="v1")
        model_deployment = ml_model_client.update_model_deployment_meta(
            model_deployment=model_deployment,
            model_meta={
                "model_lib": "xgboost",
                "model_lib_version": "1.2.0",
            },
            partial_update = True
        )
        manual_model_deployment_without_meta = ml_model_client.update_manual_model_meta(
            model_deployment=model_deployment,
            model_meta={}
        )
        ```
        """
        return self.update_manual_model_meta(model_deployment, model_meta, partial_update)

    def update_model_deployment_components(
        self,
        model_deployment: ManualModel,
        components: List[Component],
    ) -> ManualModel:
        """
        일반 모델 배포의 컴포넌트 정보를 수정합니다.
        위 메소드는 update_manual_model_components와 같은 기능을 수행합니다.

        ## Args

        - model_deployment: (`sktmls.models.ManualModel`) 일반 모델 객체
        - components: list(`sktmls.components.Component`) 수정할 컴포넌트 리스트

        ## Returns
        `sktmls.models.ManualModel`

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        component = component_client.get_component(name="test_component")
        model_deployment = ml_model_client.get_model_deployment(name="manual_test_model", version="v1")

        updated_manual_model = ml_model_client.update_model_deployment_components(
            model_deployment=model_deployment,
            components=[component],
        )
        no_component_manual_model = ml_model_client.update_model_deployment_components(
            model_deployment=model_deployment,
            components=[],
        )
        ```
        """
        return self.update_manual_model_components(model_deployment, components)

    def list_model_deployments(self, **kwargs) -> List[ManualModel]:
        """
        일반 모델 배포들을 가져옵니다.
        위 메소드는 list_manual_models와 같은 기능을 수행합니다.

        ## Args

        - kwargs: (optional) (dict) 쿼리 조건
            - id: (int) ML모델 고유 ID
            - name: (str) ML모델 이름
            - version: (str) ML모델 이름
            - query: (str) 검색 문자
            - page: (int) 페이지 번호

        ## Returns
        list(`sktmls.models.ManualModel`)

        - id: (int) ML모델 고유 ID
        - name: (str) ML모델 이름
        - model_type: (str) ML모델 타입 (`automl` | `manual`)
        - version: (str) ML모델 버전
        - description: (str) ML모델 설명
        - creator: (str) ML모델 생성 계정명
        - status: (`sktmls.models.MLModelStatus`) ML모델 상태
        - model_meta: (dict) ML모델 메타정보
        - components: list(`sktmls.components.Component`) ML모델 컴포넌트
        - table: (str) ML모델 테이블
        - model_lib: (str) ML모델 라이브러리
        - model_data: (str) ML모델 데이터
        - features: (str) ML모델에 사용된 피쳐
        - is_enabled: (bool) 활성상태 여부

        ## Example

        ```python
        manual_models = ml_model_client.list_model_deployments()
        ```
        """
        return self.list_manual_models(**kwargs)

    def start_automl_model_server(self, model: AutoMLModel) -> MLSResponse:
        """
        AutoML모델 서버를 시작합니다.

        ## Args

        - model: (`sktmls.models.AutoMLModel`) AutoML 객체

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```python
        ml_model_client.start_automl_model_server(model=model)
        ```
        """

        assert type(model) == AutoMLModel, "`model`은 `sktmls.models.AutoMLModel` 타입이어야 합니다."

        return self._request(method="POST", url=f"{MLS_MODELS_API_URL}/{model.id}/automl_model_server")

    def stop_automl_model_server(self, model: AutoMLModel) -> MLSResponse:
        """
        AutoML모델 서버를 중단합니다.

        ## Args

        - model: (`sktmls.models.AutoMLModel`) AutoML 객체

        ## Returns
        `sktmls.MLSResponse`

        ## Example

        ```python
        ml_model_client.stop_automl_model_server(model=model)
        ```
        """

        assert type(model) == AutoMLModel, "`model`은 `sktmls.models.AutoMLModel` 타입이어야 합니다."

        return self._request(method="DELETE", url=f"{MLS_MODELS_API_URL}/{model.id}/automl_model_server")

    def get_statsu_automl_model_server(self, model: AutoMLModel) -> str:
        """
        AutoML모델 서버 상태를 확인합니다.

        ## Args

        - model: (`sktmls.models.AutoMLModel`) AutoML 객체

        ## Returns
        str (`RUNNING` | `STOPPED` | `FAILED` | `UNKNOWN`)

        ## Example

        ```python
        ml_model_client.get_statsu_automl_model_server(model=model)
        ```
        """

        assert type(model) == AutoMLModel, "`model`은 `sktmls.models.AutoMLModel` 타입이어야 합니다."

        results = self._request(method="GET", url=f"{MLS_MODELS_API_URL}/{model.id}/automl_model_server").results

        return results.get("status")



import pytest

from sktmls.models.contrib import VasXcloudRuleModel


class TestVasXcloudRuleModels:
    @pytest.fixture(scope="class")
    def param(self):
        return {
            "model": VasXcloudRuleModel(
                "test_model",
                "test_version",
                [
                    "age",
                    "real_arpu_bf_m1",
                    "app_use_traffic_game",
                    "app_use_days_video_median_yn",
                    "data_use_night_ratio_median_yn",
                ],
            )
        }

    def test_000_vax_xcloud_return(self, param):
        assert param.get("model").predict([5, 55000, 10000, "Y", "Y"]) == {"items": []}

    def test_001_vax_xcloud_return(self, param):
        assert param.get("model").predict([20, 10000, 10000, "Y", "Y"]) == {"items": []}

    def test_002_vax_xcloud_return(self, param):
        assert param.get("model").predict([20, 55000, 0, "N", "N"]) == {"items": []}

    def test_003_vax_xcloud_return(self, param):
        assert param.get("model").predict([20, 55000, 10000, "Y", "Y"]) == {
            "items": [
                {
                    "id": "NA00007114",
                    "name": "게임패스 얼티밋",
                    "type": "vas_xcloud",
                    "props": {"context_id": "context_default"},
                }
            ]
        }

    def test_004_vax_xcloud_return(self, param):
        assert param.get("model").predict([20, 55000, 0, "Y", "Y"]) == {
            "items": [
                {
                    "id": "NA00007114",
                    "name": "게임패스 얼티밋",
                    "type": "vas_xcloud",
                    "props": {"context_id": "context_default"},
                }
            ]
        }

    def test_005_vax_xcloud_return(self, param):
        assert param.get("model").predict([20, 55000, 10000, "N", "Y"]) == {
            "items": [
                {
                    "id": "NA00007114",
                    "name": "게임패스 얼티밋",
                    "type": "vas_xcloud",
                    "props": {"context_id": "context_default"},
                }
            ]
        }

    def test_006_vax_xcloud_return(self, param):
        assert param.get("model").predict([20, 55000, 10000, "Y", "N"]) == {
            "items": [
                {
                    "id": "NA00007114",
                    "name": "게임패스 얼티밋",
                    "type": "vas_xcloud",
                    "props": {"context_id": "context_default"},
                }
            ]
        }



import numpy as np
from typing import Callable, Type, NamedTuple, Optional
import copy
import logging
import ConfigSpace as CS

from .debug_log import DebugLogPrinter
from .gp_profiling import GPMXNetSimpleProfiler
from .hp_ranges import HyperparameterRanges_CS
from ..datatypes.common import CandidateEvaluation, Candidate, candidate_for_print, PendingEvaluation
from ..datatypes.hp_ranges import HyperparameterRanges
from ..datatypes.tuning_job_state import TuningJobState
from ..models.gpmxnet import GPModel, GPMXNetModel
from ..models.gpmxnet_skipopt import SkipOptimizationPredicate
from ..models.gpmxnet_transformers import GPMXNetPendingCandidateStateTransformer, GPMXNetModelArgs
from ..tuning_algorithms.base_classes import LocalOptimizer, AcquisitionFunction, ScoringFunction
from ..tuning_algorithms.bo_algorithm import BayesianOptimizationAlgorithm
from ..tuning_algorithms.bo_algorithm_components import IndependentThompsonSampling
from ..tuning_algorithms.common import RandomStatefulCandidateGenerator, compute_blacklisted_candidates
from ..tuning_algorithms.default_algorithm import dictionarize_objective, DEFAULT_METRIC, DEFAULT_LOCAL_OPTIMIZER_CLASS, DEFAULT_NUM_INITIAL_CANDIDATES, DEFAULT_NUM_INITIAL_RANDOM_EVALUATIONS
from ..utils.duplicate_detector import DuplicateDetectorIdentical

logger = logging.getLogger(__name__)

GET_CONFIG_RANDOM_RETRIES = 50


class MapReward(NamedTuple):
    forward: Callable[[float], float]
    reverse: Callable[[float], float]

    def __call__(self, x: float) -> float:
        return self.forward(x)


def accumulate_profiling_record(
        cum_records: dict, profiler: GPMXNetSimpleProfiler, pick_random: bool):
    # Pull out profiling data from this block
    block_id = profiler.id_counter - 1
    curr_record = {
        r.tag: r.duration for r in profiler.records
        if r.id == block_id}
    if pick_random:
        curr_record['num_random'] = 1
        curr_record['total_all'] = curr_record['random']
    else:
        curr_record['num_model'] = 1
    # Sum up entries
    for k, v in curr_record.items():
        if k in cum_records:
            cum_records[k] += v
        else:
            cum_records[k] = v


SUPPORTED_INITIAL_SCORING = {
    'thompson_indep',
    'acq_func'}


DEFAULT_INITIAL_SCORING = 'thompson_indep'


def create_initial_candidates_scorer(
        initial_scoring: str, model: GPMXNetModel,
        acquisition_class: Type[AcquisitionFunction],
        random_state: np.random.RandomState) -> ScoringFunction:
    if initial_scoring == 'thompson_indep':
        return IndependentThompsonSampling(model, random_state=random_state)
    else:
        return acquisition_class(model)


def check_initial_candidates_scorer(initial_scoring: str) -> str:
    if initial_scoring is None:
        return DEFAULT_INITIAL_SCORING
    else:
        assert initial_scoring in SUPPORTED_INITIAL_SCORING, \
            "initial_scoring = '{}' is not supported".format(
                initial_scoring)
        return initial_scoring


class GPFIFOSearcher(object):
    """
    Supports standard GP-based hyperparameter optimization, when used with a
    FIFO scheduler.

    """
    def __init__(
            self, hp_ranges: HyperparameterRanges, random_seed: int,
            gpmodel: GPModel, model_args: GPMXNetModelArgs,
            map_reward: MapReward,
            acquisition_class: Type[AcquisitionFunction],
            init_state: TuningJobState = None,
            local_minimizer_class: Type[LocalOptimizer] = DEFAULT_LOCAL_OPTIMIZER_CLASS,
            skip_optimization: SkipOptimizationPredicate = None,
            num_initial_candidates: int = DEFAULT_NUM_INITIAL_CANDIDATES,
            num_initial_random_choices: int = DEFAULT_NUM_INITIAL_RANDOM_EVALUATIONS,
            initial_scoring: Optional[str] = None,
            profiler: Optional[GPMXNetSimpleProfiler] = None,
            first_is_default: bool = True,
            debug_log: Optional[DebugLogPrinter] = None):
        """
        Note that the GPMXNetModel is created on demand (by the state
        transformer) in get_config, along with components needed for the BO
        algorithm.

        The searcher is supposed to maximize reward, while internally, the
        criterion is minimized. map_reward maps reward to internal criterion, it
        must be strictly decreasing.

        :param hp_ranges: Configuration space without resource attribute
        :param random_seed:
        :param gpmodel: GP regression model
        :param model_args: Arguments for GPMXNet model creation
        :param map_reward: Function mapping reward to criterion to be minimized
        :param acquisition_class: Type for acquisition function
        :param init_state: TuningJobState to start from (default is empty)
        :param local_minimizer_class: Type for local minimizer
        :param skip_optimization: Predicate, see
            GPMXNetPendingCandidateStateTransformer
        :param num_initial_candidates: See BayesianOptimizationAlgorithm
        :param num_initial_random_choices: Configs are sampled at random until
            this many candidates received label feedback
        :param initial_scoring: Scoring function to rank initial candidates.
            Default: thompson_indep (independent Thompson sampling)
        :param profiler: If given, HPO computations are profiled
        :param first_is_default: If true, the first config to be evaluated is
            the default of the search space. Otherwise, the first is sampled
            at random
        :param debug_log: DebugLogPrinter for debug logging (optional)

        """
        self.hp_ranges = hp_ranges
        self.random_seed = random_seed
        self.num_initial_candidates = num_initial_candidates
        self.num_initial_random_choices = num_initial_random_choices
        self.map_reward = map_reward
        self.local_minimizer_class = local_minimizer_class
        self.acquisition_class = acquisition_class
        self.debug_log = debug_log
        self.initial_scoring = check_initial_candidates_scorer(initial_scoring)
        # Create state transformer
        # Initial state is empty (note that the state is mutable)
        if init_state is None:
            init_state = TuningJobState(
                hp_ranges=hp_ranges,
                candidate_evaluations=[],
                failed_candidates=[],
                pending_evaluations=[])
        else:
            assert hp_ranges is init_state.hp_ranges, \
                "hp_ranges and init_state.hp_ranges must be same object"
        self.state_transformer = GPMXNetPendingCandidateStateTransformer(
            gpmodel=gpmodel,
            init_state=init_state,
            model_args=model_args,
            skip_optimization=skip_optimization,
            profiler=profiler,
            debug_log=debug_log)
        self.random_state = np.random.RandomState(random_seed)
        self.random_generator = RandomStatefulCandidateGenerator(
            hp_ranges, random_state=self.random_state)
        self.profiler = profiler
        self.do_profile = (profiler is not None)
        self.first_is_default = first_is_default
        if first_is_default:
            assert isinstance(hp_ranges, HyperparameterRanges_CS), \
                "If first_is_default, must have hp_ranges of HyperparameterRanges_CS type"
        if debug_log is not None:
            assert isinstance(hp_ranges, HyperparameterRanges_CS), \
                "If debug_log is given, must have hp_ranges of HyperparameterRanges_CS type"
        # Sums up profiling records across all get_config calls
        self._profile_record = dict()
        if debug_log is not None:
            deb_msg = "[GPFIFOSearcher.__init__]\n"
            deb_msg += ("- acquisition_class = {}\n".format(acquisition_class))
            deb_msg += ("- local_minimizer_class = {}\n".format(local_minimizer_class))
            deb_msg += ("- num_initial_candidates = {}\n".format(num_initial_candidates))
            deb_msg += ("- num_initial_random_choices = {}\n".format(num_initial_random_choices))
            deb_msg += ("- initial_scoring = {}\n".format(self.initial_scoring))
            deb_msg += ("- first_is_default = {}".format(first_is_default))
            logger.info(deb_msg)

    def update(self, config: Candidate, reward: float):
        """
        Registers new datapoint at config, with reward reward.
        Note that in general, config should previously have been registered as
        pending (register_pending). If so, it is switched from pending
        to labeled. If not, it is considered directly labeled.

        :param config:
        :param reward:
        """
        crit_val = self.map_reward(reward)
        self.state_transformer.label_candidate(CandidateEvaluation(
            candidate=copy.deepcopy(config),
            metrics=dictionarize_objective(crit_val)))
        if self.debug_log is not None:
            config_id = self.debug_log.config_id(config)
            msg = "Update for config_id {}: reward = {}, crit_val = {}".format(
                config_id, reward, crit_val)
            logger.info(msg)

    def register_pending(self, config: Candidate):
        """
        Registers config as pending. This means the corresponding evaluation
        task is running. Once it finishes, update is called for config.

        """
        # It is OK for the candidate already to be registered as pending, in
        # which case we do nothing
        state = self.state_transformer.state
        if config not in state.pending_candidates:
            if config in (x.candidate for x in state.candidate_evaluations):
                evals = state.candidate_evaluations
                num_labeled = len(evals)
                pos_cand = next(
                    i for i, x in enumerate(evals) if x.candidate == config)
                error_msg = """
                This configuration is already registered as labeled:
                   Position of labeled candidate: {} of {}
                   Label value: {}
                """.format(
                    pos_cand, num_labeled,
                    evals[pos_cand].metrics[DEFAULT_METRIC])
                assert False, error_msg
            self.state_transformer.append_candidate(config)

    def get_config(self) -> Candidate:
        """
        Runs Bayesian optimization in order to suggest the next config to evaluate.

        :return: Next config to evaluate at
        """
        state = self.state_transformer.state
        if self.do_profile:
            fit_hyperparams = not self.state_transformer.skip_optimization(
                state)
            self.profiler.set_state(state, fit_hyperparams)
        blacklisted_candidates = compute_blacklisted_candidates(state)
        pick_random = (len(blacklisted_candidates) < self.num_initial_random_choices) or \
            (not state.candidate_evaluations)
        if self.debug_log is not None:
            self.debug_log.start_get_config('random' if pick_random else 'BO')
        if pick_random:
            config = None
            if self.first_is_default and (not blacklisted_candidates):
                # Use default configuration if there is one specified
                default_config = self.hp_ranges.config_space.get_default_configuration()
                if default_config and len(default_config.get_dictionary()) > 0:
                    config = default_config
                    if self.debug_log is not None:
                        logger.info("Start with default config:\n{}".format(
                            candidate_for_print(config)))
            if config is None:
                if self.do_profile:
                    self.profiler.start('random')
                for _ in range(GET_CONFIG_RANDOM_RETRIES):
                    _config = self.hp_ranges.random_candidate(self.random_state)
                    if _config not in blacklisted_candidates:
                        config = _config
                        break
                if config is None:
                    raise AssertionError(
                        "Failed to sample a configuration not already chosen "
                        "before. Maybe there are no free configurations left? "
                        "The blacklist size is {}".format(len(blacklisted_candidates)))
                if self.do_profile:
                    self.profiler.stop('random')
        else:
            # Obtain current GPMXNetModel from state transformer. Based on
            # this, the BO algorithm components can be constructed
            state = self.state_transformer.state
            if self.do_profile:
                self.profiler.start('total_all')
                self.profiler.start('total_update')
            # Note: Asking for the model triggers the posterior computation
            model = self.state_transformer.model()
            if self.do_profile:
                self.profiler.stop('total_update')
            # Create BO algorithm
            initial_candidates_scorer = create_initial_candidates_scorer(
                self.initial_scoring, model, self.acquisition_class,
                self.random_state)
            local_optimizer = self.local_minimizer_class(
                state, model, self.acquisition_class)
            # Make sure not to use the same random seed for each call:
            #random_seed = compute_random_seed({'0': state}, self.random_seed)
            bo_algorithm = BayesianOptimizationAlgorithm(
                initial_candidates_generator=self.random_generator,
                initial_candidates_scorer=initial_candidates_scorer,
                num_initial_candidates=self.num_initial_candidates,
                local_optimizer=local_optimizer,
                pending_candidate_state_transformer=None,
                blacklisted_candidates=blacklisted_candidates,
                num_requested_candidates=1,
                greedy_batch_selection=False,
                duplicate_detector=DuplicateDetectorIdentical(),
                profiler=self.profiler,
                sample_unique_candidates=False,
                debug_log=self.debug_log)
            # Next candidate decision
            if self.do_profile:
                self.profiler.start('total_nextcand')
            _config = bo_algorithm.next_candidates()
            if len(_config) == 0:
                raise AssertionError(
                    "Failed to find a configuration not already chosen "
                    "before. Maybe there are no free configurations left? "
                    "The blacklist size is {}".format(len(blacklisted_candidates)))
            config = _config[0]
            if self.do_profile:
                self.profiler.stop('total_nextcand')
            if self.do_profile:
                self.profiler.stop('total_all')

        if self.debug_log is not None:
            self.debug_log.set_final_config(config)
            # All get_config debug log info is only written here
            self.debug_log.write_block()
        if self.do_profile:
            self.profiler.clear()
            # Pull out profiling data from this block, add to _profile_record
            accumulate_profiling_record(
                self._profile_record, self.profiler, pick_random)

        return config

    def evaluation_failed(self, config: Candidate):
        # Remove pending candidate
        self.state_transformer.drop_candidate(config)
        # Mark config as failed (which means it will be blacklisted in
        # future get_config calls)
        self.state_transformer.mark_candidate_failed(config)

    def dataset_size(self):
        return len(self.state_transformer.state.candidate_evaluations)

    def cumulative_profile_record(self):
        """
        If profiling is activated, we sum up the profiling blocks for each
        call of get_config and return it as dict. See get_config for what
        is recorded:
        - num_random: Number of get_config calls with random selection
        - num_model: Number of get_config calls with model-based selection
        - total_all: Sum of total times for all get_config calls
        """
        return self._profile_record

    def get_params(self):
        """
        Note: Once MCMC is supported, this method will have to be refactored.

        :return: Dictionary with current hyperparameter values
        """
        return self.state_transformer.get_params()

    def set_params(self, param_dict):
        self.state_transformer.set_params(param_dict)

    def get_state(self):
        """
        The mutable state consists of the GP model parameters, the
        TuningJobState, and the skip_optimization predicate (which can have a
        mutable state).
        We assume that skip_optimization can be pickled.

        """
        state = {
            'model_params': self.get_params(),
            'state': encode_state(self.state_transformer.state),
            'skip_optimization': self.state_transformer.skip_optimization,
            'random_state': self.random_state}
        if self.debug_log is not None:
            state['debug_log'] = self.debug_log.get_mutable_state()
        return state

    def clone_from_state(self, state):
        # Create clone with mutable state taken from 'state'
        model_args = self.state_transformer._model_args
        init_state = decode_state(state['state'], self.hp_ranges)
        skip_optimization = state['skip_optimization']
        new_searcher = GPFIFOSearcher(
            hp_ranges=self.hp_ranges,
            random_seed=self.random_seed,
            gpmodel=self.state_transformer._gpmodel,
            model_args=model_args,
            map_reward=self.map_reward,
            acquisition_class=self.acquisition_class,
            init_state=init_state,
            local_minimizer_class=self.local_minimizer_class,
            skip_optimization=skip_optimization,
            num_initial_candidates=self.num_initial_candidates,
            num_initial_random_choices=self.num_initial_random_choices,
            initial_scoring=self.initial_scoring,
            profiler=self.profiler,
            first_is_default=self.first_is_default,
            debug_log=self.debug_log)
        new_searcher.state_transformer.set_params(state['model_params'])
        new_searcher.random_state = state['random_state']
        new_searcher.random_generator.random_state = \
            state['random_state']
        if self.debug_log and 'debug_log' in state:
            new_searcher.debug_log.set_mutable_state(state['debug_log'])
        # Invalidate self (must not be used afterwards)
        self.state_transformer = None
        return new_searcher


def encode_state(state: TuningJobState) -> dict:
    assert isinstance(state.hp_ranges, HyperparameterRanges_CS), \
        "Must have hp_ranges of HyperparameterRanges_CS type"
    candidate_evaluations = [
        {'candidate': eval.candidate.get_dictionary(),
         'metrics': eval.metrics}
        for eval in state.candidate_evaluations]
    failed_candidates = [x.get_dictionary() for x in state.failed_candidates]
    pending_evaluations = [
        eval.candidate.get_dictionary() for eval in state.pending_evaluations]
    return {
        'candidate_evaluations': candidate_evaluations,
        'failed_candidates': failed_candidates,
        'pending_evaluations': pending_evaluations}


def decode_state(enc_state: dict, hp_ranges: HyperparameterRanges_CS) \
        -> TuningJobState:
    assert isinstance(hp_ranges, HyperparameterRanges_CS), \
        "Must have hp_ranges of HyperparameterRanges_CS type"
    config_space = hp_ranges.config_space

    def to_cs(x):
        return CS.Configuration(config_space, values=x)

    candidate_evaluations = [
        CandidateEvaluation(to_cs(x['candidate']), x['metrics'])
        for x in enc_state['candidate_evaluations']]
    failed_candidates = [to_cs(x) for x in enc_state['failed_candidates']]
    pending_evaluations = [
        PendingEvaluation(to_cs(x)) for x in enc_state['pending_evaluations']]
    return TuningJobState(
        hp_ranges=hp_ranges,
        candidate_evaluations=candidate_evaluations,
        failed_candidates=failed_candidates,
        pending_evaluations=pending_evaluations)


def map_reward(const=1.0) -> MapReward:
    """
    Factory for map_reward argument in GPMultiFidelitySearcher.
    """
    def const_minus_x(x):
        return const - x

    return MapReward(forward=const_minus_x, reverse=const_minus_x)



import os
from typing import Any, Dict

import requests

from sktmls import MLSENV, MLSRuntimeENV
from sktmls.config import Config


class MLSConversionTrackingAPIClient:
    """
    MLS Conversion Tracking API를 호출할 수 있는 클라이언트입니다.

    모든 호출은 테스트용 호출로 처리됩니다.

    MMS, EDD 환경은 지원하지 않습니다.
    """

    def __init__(
        self,
        client_id: str,
        apikey: str,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV = None,
    ):
        """
        ## Args

        - client_id: (str) 호출 클라이언트 (`tw`, `twd`, `netcrm` 등)
        - apikey: (str) 클라이언트의 API Key
        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env

        ## Returns
        `sktmls.apis.MLSConversionTrackingAPIClient`

        ## Example

        ```python
        conversion_tracking_api_client = MLSConversionTrackingAPIClient(client_id="tw", apikey="test_apikey", env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE)
        ```
        """
        if env:
            assert env in MLSENV.list_items(), "유효하지 않은 MLS 환경입니다."
            self.__env = env
        elif os.environ.get("MLS_ENV"):
            self.__env = MLSENV[os.environ["MLS_ENV"]]
        elif os.environ.get("AWS_ENV"):
            self.__env = MLSENV[os.environ["AWS_ENV"]]
        else:
            self.__env = MLSENV.STG

        if runtime_env:
            assert runtime_env in MLSRuntimeENV.list_items(), "유효하지 않은 런타임 환경입니다."
            self.__runtime_env = runtime_env
        elif os.environ.get("MLS_RUNTINE_ENV"):
            self.__runtime_env = MLSRuntimeENV[os.environ["MLS_RUNTINE_ENV"]]
        else:
            __HOSTNAME = os.environ.get("HOSTNAME", "").lower()
            if __HOSTNAME.startswith("bdp-dmi"):
                self.__runtime_env = MLSRuntimeENV.YE
            elif __HOSTNAME.startswith("vm-skt"):
                self.__runtime_env = MLSRuntimeENV.EDD
            else:
                self.__runtime_env = MLSRuntimeENV.LOCAL

        assert runtime_env not in [MLSRuntimeENV.EDD, MLSRuntimeENV.MMS], "EDD, MMS 환경은 지원하지 않습니다."
        assert client_id and apikey, "`client_id`와 `apikey`가 전달되어야 합니다."
        self.__client_id = client_id
        self.__apikey = apikey

        self.config = Config(self.__runtime_env.value)

    def get_env(self) -> MLSENV:
        return self.__env

    def get_runtime_env(self) -> MLSRuntimeENV:
        return self.__runtime_env

    def request(
        self, user_id: str, channel_id: str, conversion_type: str, process_id: str, item_id: str
    ) -> Dict[str, Any]:
        """
        MLS Conversion Tracking API를 호출합니다.

        ## Args

        - user_id: (str) 추천 요청의 user ID (해시된 서비스관리번호)
        - channel_id: (str) 추천 요청 또는 응답의 channel ID
        - conversion_type: (str) 이벤트 타입(`impression`|`click`|`detail_view`|`like`|`dislike`)
        - process_id: (str) 추천 결과 응답의 process ID
        - item_id : (str) 추천 결과 응답의 item_id

        ## Returns
        dict

        ## Example

        ```python
        conversion_tracking_dict = conversion_tracking_api_client.request(
            user_id=hashlib.sha256("1234567890".encode()).hexdigest(),
            channel_id="sample_channel",
            conversion_type="impression",
            process_id="sample_process",
            item_id="sample_item",
        )
        ```
        """

        client_id = self.__client_id
        headers = {"apikey": self.__apikey, "User-Id-Want-Digest": "no"}
        params = {"mode": "test"}
        data = {
            "user_id": user_id,
            "channel_id": channel_id,
            "process_id": process_id,
            "item_id": item_id,
        }
        response = requests.post(
            url=f"{self.config.MLS_CONVERSION_TRACKING_API_URL[self.__env.value]}/v1/{client_id}/{conversion_type}",
            headers=headers,
            params=params,
            json=data,
        )

        return response.content.decode("utf-8")



import logging

from ...constants import BINARY, MULTICLASS, REGRESSION
from ...models.tabular_nn.tabular_nn_model import TabularNeuralNetModel
from ....metrics import mean_squared_error, log_loss
from ...models.rf.rf_model import RFModel
from .presets import get_preset_models, get_preset_models_softclass

logger = logging.getLogger(__name__)

# Higher values indicate higher priority, priority dictates the order models are trained (which matters if time runs out).
DEFAULT_DISTILL_PRIORITY = dict(
    GBM=100,
    NN=90,
    CAT=80,
    RF=70,
    custom=0,
)


def get_preset_models_distillation(path, problem_type, eval_metric, hyperparameters, stopping_metric=None, num_classes=None,
                                   hyperparameter_tune=False, distill_level=0, name_suffix='_DSTL'):
    if problem_type == MULTICLASS:
        models = get_preset_models_softclass(path=path, num_classes=num_classes, hyperparameters=hyperparameters,
                                             hyperparameter_tune=hyperparameter_tune, name_suffix=name_suffix)
    elif problem_type == BINARY:  # convert to regression in distillation
        eval_metric = mean_squared_error
        stopping_metric = mean_squared_error
        # Constrain output-range of NN:
        nn_outputrange = {'y_range': (0.0,1.0), 'y_range_extend': 0.0}
        if 'NN' in hyperparameters:
            nn_hyperparameters = hyperparameters['NN']
        elif 'default' in hyperparameters and 'NN' in hyperparameters['default']:
            nn_hyperparameters = hyperparameters['default']['NN']
        else:
            nn_hyperparameters = None
        if isinstance(nn_hyperparameters, list):
            for i in range(len(nn_hyperparameters)):
                nn_hyperparameters[i].update(nn_outputrange)
        elif nn_hyperparameters is not None:
            nn_hyperparameters.update(nn_outputrange)
        if 'NN' in hyperparameters:
            hyperparameters['NN'] = nn_hyperparameters
        elif 'default' in hyperparameters and 'NN' in hyperparameters['default']:
            hyperparameters['default']['NN'] = nn_hyperparameters
        # Swap RF criterion for MSE:
        rf_newparams = {'criterion': 'mse', 'AG_args': {'name_suffix': 'MSE'}}
        if 'RF' in hyperparameters:
            rf_hyperparameters = hyperparameters['RF']
        elif 'default' in hyperparameters and 'RF' in hyperparameters['default']:
            rf_hyperparameters = hyperparameters['default']['RF']
        else:
            rf_hyperparameters = None
        if isinstance(rf_hyperparameters, list):
            for i in range(len(rf_hyperparameters)):
                rf_hyperparameters[i].update(rf_newparams)
            rf_hyperparameters = [j for n, j in enumerate(rf_hyperparameters) if j not in rf_hyperparameters[(n+1):]]  # Remove duplicates which may arise after overwriting criterion
        elif rf_hyperparameters is not None:
            rf_hyperparameters.update(rf_newparams)
        if 'RF' in hyperparameters:
            hyperparameters['RF'] = rf_hyperparameters
        elif 'default' in hyperparameters and 'RF' in hyperparameters['default']:
            hyperparameters['default']['RF'] = rf_hyperparameters

    if problem_type == REGRESSION or problem_type == BINARY:
        models = get_preset_models(path=path, problem_type=REGRESSION, eval_metric=eval_metric, stopping_metric=stopping_metric,
                                   hyperparameters=hyperparameters, hyperparameter_tune=hyperparameter_tune, name_suffix=name_suffix, default_priorities=DEFAULT_DISTILL_PRIORITY)

    if problem_type in [MULTICLASS, BINARY]:
        for model in models:
            model.normalize_pred_probas = True
            model.name = model.name.replace('Regressor', 'Classifier')  # conceal from user that model may actually be a regressor.
            model.name = model.name.replace('SoftClassifier', 'Classifier')

    logger.log(20, f"Distilling with each of these student models: {[model.name for model in models]}")
    return models



from typing import Any, Dict, List

import random

from sktmls.models import MLSRuleModel


class TwRandomGreetingRuleModel(MLSRuleModel):
    """
    MLS 모델 레지스트리에 등록되는 Rule 기반 클래스입니다.

    TW 그리팅 추천 모델에 특화된 모델이며, 2개의 그리팅 타입(이미지 / 텍스트) 및 1개의 그리팅 랭킹을 사용합니다.
    """

    def __init__(
        self,
        model_name: str,
        model_version: str,
        image_type_ids: List[str],
        text_type_ids: List[str],
        greeting_ids: List[str],
        greeting_ids_shuffle: bool = False,
    ):
        """
        ## Args

        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - image_type_ids: (list) 티월드그리팅 이미지 타입
        - text_type_ids: (list) 티월드그리팅 텍스트 타입
        - greeting_ids: (list) 티월드그리팅 랭킹
        - greeting_ids_shuffle: (bool) 티월드그리팅 랭킹 Shuffle 여부

        ## Example

        ```python
        random_greeting_rule_model_v1 = TwRandomGreetingRuleModel(
            model_name="random_greeting_rule_model",
            model_version="v1",
            image_type_ids=["A", "B"],
            text_type_ids=["A", "B", "C"],
            greeting_ids=["C1", "C2", "C3", "C4", "C5"],
            greeting_ids_shuffle=True
        )

        result = random_greeting_rule_model_v1.predict(None)
        ```
        """
        assert isinstance(image_type_ids, list) and image_type_ids, "`image_type_ids`는 list 형식이어야 합니다."
        assert isinstance(text_type_ids, list) and text_type_ids, "`text_type_ids`는 list 형식이어야 합니다."
        assert isinstance(greeting_ids, list) and greeting_ids, "`greeting_ids`는 list 형식이어야 합니다."

        super().__init__(model_name, model_version, ["user_id"])

        self.image_type_ids = image_type_ids
        self.text_type_ids = text_type_ids
        self.greeting_ids = greeting_ids
        self.greeting_ids_shuffle = greeting_ids_shuffle

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        if self.greeting_ids_shuffle:
            random.shuffle(self.greeting_ids)

        return {
            "items": [
                {
                    "id": "tw_greeting_image",
                    "name": "티월드그리팅이미지타입",
                    "type": "tw_greeting",
                    "props": {"bucket": random.choice(self.image_type_ids)},
                },
                {
                    "id": "tw_greeting_text",
                    "name": "티월드그리팅텍스트타입",
                    "type": "tw_greeting",
                    "props": {"bucket": random.choice(self.text_type_ids)},
                },
                {
                    "id": "tw_greeting_ranking",
                    "name": "티월드그리팅랭킹",
                    "type": "tw_greeting",
                    "props": {"ranking": self.greeting_ids},
                },
            ]
        }



from typing import Any, Dict, List

import re
from functools import reduce

import numpy as np

from sktmls.models import MLSLightGBMModel, MLSModelError
from sktmls.meta_tables import MetaTable


class LightGBMDeviceModel(MLSLightGBMModel):
    """
    MLS 모델 레지스트리에 등록되는 LightGBM 기반 클래스입니다.

    주어진 `features`리스트를 이용해 prediction 후 스코어 상위 3개 단말그룹으로 필터링 합니다.
    이후 `id`, `name`, `score`, `props`, `type`을 반환하는 모델입니다.
    """

    def __init__(
        self,
        model,
        model_name: str,
        model_version: str,
        model_features: List[str],
        default_model_feature_values: List[Any],
        label_indices: Dict[str, int],
        product_classes: Dict[str, Any],
        products: Dict[str, Any],
        device_meta: MetaTable,
        conversion_formulas: Dict[str, Dict[str, Any]] = dict(),
        emb_features: List[str] = [],
        default_emb_feature_values: List[List[Any]] = [],
        emb_feature_indices: Dict[str, List[int]] = [],
        default_context_value: str = "context_default",
        num_pick: int = 3,
        high_rank_word: List[str] = ["Plus", "플러스", "울트라", "Ultra", "Max", "맥스", "Pro", "프로"],
    ):
        """
        ## Args

        - model: LightGBM으로 학습한 모델 객체
        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - model_features: (list(str)) non-embedding 피쳐 리스트(필수피쳐: "eqp_mdl_cd", "age")
        - default_model_feature_values: (list) 피쳐 기본값 리스트
        - label_indices: (dict) 라벨 인덱스 값
        - product_classes: (dict) 상품 클래스 값
        - products: (dict) 컨텍스트 값
        - device_meta: (`sktmls.meta_tables.MetaTable`) 단말코드그룹 값
        - conversion_formulas: (optional) (dict(dict)) Feature conversion에 사용할 조건 dict. 형식은 아래 Example 참조 (기본값: {})
        - emb_features: (optional) (list(str)) 임베딩 피쳐 이름 리스트 (기본값: [])
        - default_emb_feature_values: (optional) (list(list)) 임베딩 피쳐들의 기본값 리스트 (리스트의 리스트 형식) (기본값: [])
        - emb_feature_indices: (optional) (list(list)) 각 임베딩 피쳐의 세부 사용 피쳐 인덱스 리스트 (기본값: [])
        - default_context_value: (optional) (str) 기본 컨텍스트 값 (기본값: "context_default")
        - num_pick: (optional) (int) 스코어 상위 N개 추출
        - high_rank_word: (optional) (list(str)) 플래그쉽 모델 명칭 (기본값 : ["Plus", "플러스", "울트라", "Ultra", "Max", "맥스", "Pro", "프로"])

        ## Example

        ```python
        model_features = ["feature1", "feature2", "feature3"]
        default_model_feature_values = [
            dimension_client.get_dimension(dimension_type="user", name=feature).default for feature in model_features
        ]

        label_indices = {
            "삼성전자_120만원": 0,
            "삼성전자_160만원": 1,
            "삼성전자_60만원": 2,
            "Apple_120만원": 3,
            "Apple_160만원": 4,
            "Apple_60만원": 5,
            "LG전자_120만원": 6,
            "LG전자_160만원": 7,
            "LG전자_60만원": 8,
            "기타_120만원": 9,
            "기타_160만원": 10,
            "기타_60만원": 11,
        }

        product_classes = {
            "Apple_120만원": [
                {
                    "product_grp_id": "000004352",
                    "product_grp_nm": "iPhone SE (2020)",
                    "mfact_nm": "Apple",
                    "eqp_mdl_cd": "A23F",
                    "rep_eqp_mdl_cd": "A23F",
                    "rep_eqp_yn": "Y",
                    "color_hex": "1F2120",
                    "color_nm": "블랙",
                    "network_type": "4G",
                    "device_weight": 0.02356698,
                },
                {
                    "product_grp_id": "000003972",
                    "product_grp_nm": "iPhone 11",
                    "mfact_nm": "Apple",
                    "eqp_mdl_cd": "A1G6",
                    "rep_eqp_mdl_cd": "A1G6",
                    "rep_eqp_yn": "Y",
                    "color_hex": "1F2120",
                    "color_nm": "블랙",
                    "network_type": "4G",
                    "device_weight": 0.01638297,
                },
            ],
            "Apple_160만원": [
                {
                    "product_grp_id": "000003973",
                    "product_grp_nm": "iPhone 11 Pro",
                    "mfact_nm": "Apple",
                    "eqp_mdl_cd": "A1GR",
                    "rep_eqp_mdl_cd": "A1GR",
                    "rep_eqp_yn": "Y",
                    "color_hex": "52514F",
                    "color_nm": "스페이스 그레이",
                    "network_type": "4G",
                    "device_weight": 0.00767129,
                }
            ],
        }

        products = {
            "000000332": {
                "name": "iPhone 6",
                "type": "device",
                "context_features": ["context_default"],
                "random_context": False,
            },
            "000000952": {
                "name": "iPhone 6s",
                "type": "device",
                "context_features": ["context_default"],
                "random_context": False,
            },
            "000001153": {
                "name": "갤럭시 S7 엣지",
                "type": "device",
                "context_features": ["context_popular_apple_device"],
                "random_context": False,
            },
            "000001572": {
                "name": "BlackBerry PRIV",
                "type": "device",
                "context_features": ["context_default"],
                "random_context": False,
            },
            "000001613": {"name": "준3", "type": "device", "context_features": ["context_default"], "random_context": False},
            "000001672": {
                "name": "iPhone 7",
                "type": "device",
                "context_features": ["context_popular_apple_device"],
                "random_context": False,
            },
        }

        device_meta = metatable_client.get_meta_table(name="eqp_mdl_cd_meta")

        conversion_formulas = {
            "additional_svc_allcare_scrb_type": {"map": {"free": 1, "paid": 2, "N/A": 0}, "default": 0},
            "additional_svc_ansim_option_scrb_type": {"map": {"free": 1, "paid": 2, "N/A": 0}, "default": 0},
            "additional_svc_flo_scrb_type": {"map": {"free": 1, "paid": 2, "N/A": 0}, "default": 0},
            "additional_svc_melon_scrb_type": {"map": {"free": 1, "paid": 2, "N/A": 0}, "default": 0},
            "additional_svc_oksusu_scrb_type": {"map": {"free": 1, "paid": 2, "N/A": 0}, "default": 0},
            "additional_svc_pooq_scrb_type": {"map": {"free": 1, "paid": 2, "N/A": 0}, "default": 0},
            "age_band": {"map": {"kids": 1, "ting": 2, "young": 3, "tplan": 4, "senior": 5}, "default": 4},
            "channel": {"map": {"T월드": 1, "고객센터": 2, "오프라인": 3}, "default": 0},
            "family_comb_type": {"map": {"wire": 1, "wless": 2}, "default": 0},
            "main_channel_eqp_buy": {"map": {"D": 1, "S": 2}, "default": 0},
            "mbr_card_gr_cd": {"map": {"S": 1, "G": 2, "V": 3}, "default": 0},
            "mng_nice_cb_grd": {
                "map": {"01": 1, "02": 2, "03": 3, "04": 4, "05": 5, "06": 6, "07": 7, "08": 8, "09": 9, "10": 10},
                "default": -1,
            },
            "prefer_device_price": {"map": {"M": 1, "H": 2}, "default": 0},
            "prefer_latest_device": {"map": {"M": 1, "H": 2, "R": 3}, "default": 0},
            "prefer_device_manufacturer": {"map": {"lg": 1, "apple": 2, "samsung": 3}, "default": 0},
            "tmap_freq_dest_residence": {"map": {"apart": 1, "complex_redisence": 2, "villa": 3, "officetel": 4}, "default": 0},
        }

        my_model_v1 = LightGBMDeviceModel(
            model=lightgbm_model,
            model_name="my_model",
            model_version="v1",
            model_features=["eqp_mdl_cd", "age"] + model_features,
            default_model_feature_values=default_model_feature_values,
            label_indices=label_indices,
            product_classes=product_classes,
            products=products,
            device_meta=device_meta,
            conversion_formulas=conversion_formulas,
            emb_features=["embedding_vector"],
            default_emb_feature_values=[[0.0] * 64],
            emb_feature_indices=[[0, 1]],
        )

        result = my_model_v1.predict(["eqp_mdl_cd", "value_1", "value_2", "value_3", [[0.1, 0.2]]])

        ```
        """

        feature_indices_non_emb = [i for i, feature in enumerate(model.feature_name()) if not feature.startswith("emb")]
        feature_indices_emb = [i for i, feature in enumerate(model.feature_name()) if feature.startswith("emb")]
        if feature_indices_emb:
            assert (
                feature_indices_non_emb[-1] < feature_indices_emb[0]
            ), "`model`학습에 사용된 feature 중 embedding 피쳐는 반드시 non-embedding 피쳐 뒤에 있어야 합니다."

        assert isinstance(model_features, list), "`model_features`는 list타입이어야 합니다."
        assert model_features[0] == "eqp_mdl_cd" and model_features[1] == "age", "`eqp_mdl_cd`, `age`는 필수 피쳐 입니다."
        assert isinstance(default_model_feature_values, list), "`default_model_feature_values`는 list타입이어야 합니다."
        assert len(model_features) - 2 == len(
            default_model_feature_values
        ), "`model_features`(필수 피쳐제외)와 `default_model_feature_values`의 길이가 다릅니다."
        assert None not in default_model_feature_values, "`default_model_feature_values`에 None이 포함되어 있습니다."

        assert isinstance(emb_features, list), "`emb_features`는 리스트 형식이어야 합니다."
        assert isinstance(default_emb_feature_values, list), "`default_emb_feature_values`는 리스트 형식이어야 합니다."
        assert None not in default_emb_feature_values, "`default_emb_feature_values`에 None이 포함되어 있습니다."
        assert isinstance(emb_feature_indices, list), "`emb_feature_indices`는 리스트 형식이어야 합니다."
        assert None not in emb_feature_indices, "`emb_feature_indices`에 None이 포함되어 있습니다."
        assert (
            len(emb_features) == len(default_emb_feature_values) == len(emb_feature_indices)
        ), "`emb_features`, `default_emb_feature_values`, `emb_feature_indices`의 길이가 다릅니다."

        for i, indices in enumerate(emb_feature_indices):
            assert isinstance(indices, list) and isinstance(
                default_emb_feature_values[i], list
            ), "`emb_feature_indices`와 `default_emb_feature_values` 내 모든 element는 리스트 형식이어야 합니다."
            assert len(indices) <= len(
                default_emb_feature_values[i]
            ), "`emb_feature_indices` 내 element 길이가 해당 임베딩 피쳐 길이보다 깁니다."
            assert None not in default_emb_feature_values[i], "`default_emb_feature_values` 내 element에 None이 포함되어 있습니다."
            for idx in indices:
                assert isinstance(idx, int), "`emb_feature_indices` 내 모든 인덱스 값은 int 형식이어야 합니다."
                assert idx < len(default_emb_feature_values[i]), "`emb_feature_indices` 내 인덱스가 임베딩 피쳐 전체 길이보다 깁니다."

        assert isinstance(label_indices, dict) and len(label_indices) > 0, "`label_indices`는 필수 입력 값 입니다."
        assert isinstance(product_classes, dict) and len(product_classes) > 0, "`product_classes`는 필수 입력 값 입니다."
        assert isinstance(products, dict) and len(products) > 0, "`products`는 필수 입력 값 입니다."
        assert isinstance(high_rank_word, list), "`high_rank_word`리스트 타입이어야 합니다."

        assert isinstance(device_meta, MetaTable) and len(device_meta.items) > 0, "`device_meta`는 필수 입력 값 입니다."

        assert isinstance(conversion_formulas, dict), "`conversion_formulas`는 dict 형식이어야 합니다."
        for k, v in conversion_formulas.items():
            assert isinstance(k, str), "`conversion_formulas`의 키는 문자열이어야 합니다."
            assert isinstance(v, dict), "`conversion_formulas`의 값은 dict 형식이어야 합니다."
            assert "map" in v or "conditions" in v, "`conversion_formulas`의 값에 map 또는 conditions 키가 포함되어야 합니다."
            assert "default" in v, "`conversion_formulas`의 값에 default 키가 포함되어야 합니다."

        for device_meta_item in device_meta.items:
            device_meta_item_name = device_meta_item.get("name")
            device_meta_item_value = device_meta_item.get("values")

            assert (
                device_meta_item_value and device_meta_item_name
            ), "`device_meta`의 `item` 항목 중 `name`과 `values`는 필수입니다."
            assert isinstance(device_meta_item_name, str), "`device_meta.item`의 `name`은 str 타입이어야 합니다."
            assert isinstance(device_meta_item_value, dict), "`device_meta.item`의 `values`은 dict 타입이어야 합니다."
            assert {
                "product_grp_id",
                "product_grp_nm",
                "mfact_nm",
            } <= device_meta_item_value.keys(), (
                "`device_meta.item`의 `values`는 `product_grp_id`, `product_grp_nm`, `mfact_nm`키를 포함하여야 합니다."
            )

        super().__init__(model, model_name, model_version, model_features + emb_features)

        self.default_model_feature_values = default_model_feature_values
        self.index_labels = {v: k for k, v in label_indices.items() if k in product_classes}
        self.product_classes = product_classes
        self.products = products
        self.device_meta = {meta["name"]: meta["values"] for meta in device_meta.items}
        self.conversion_formulas = conversion_formulas
        self.default_emb_feature_values = default_emb_feature_values
        self.emb_feature_indices = emb_feature_indices
        self.default_context_value = default_context_value
        self.num_pick = num_pick
        self.high_rank_word = high_rank_word

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        # 1. 전처리
        model_x = self._preprocess(x[2:])

        # 2. ML Prediction
        y = self._ml_predict(model_x)

        # 3. 후처리
        items = self._postprocess(y, x[0], x[1])

        return {"items": items}

    def _preprocess(self, x: List[Any]) -> List[Any]:
        if len(self.features) - 2 != len(x):
            raise MLSModelError("`x`와 `features`(첫 두개(`eqp_mdl_cd`, `age`)제외)의 길이가 다릅니다.")

        # - 1.1 피쳐 나누기
        emb_x = x[len(x) - len(self.default_emb_feature_values) :]
        for i, emb_f in enumerate(emb_x):
            if not emb_f:
                emb_x[i] = self.default_emb_feature_values[i]
                continue
            if not isinstance(emb_f, list):
                raise MLSModelError("LightGBMDeviceModel: emb_x는 list 타입이어야 합니다.")
            if len(emb_f) != len(self.default_emb_feature_values[i]):
                raise MLSModelError("LightGBMDeviceModel: emb_x와 `default_emb_feature_values`의 임베딩 피쳐 길이가 다릅니다.")
            if len(emb_f) < len(self.emb_feature_indices[i]):
                raise MLSModelError("LightGBMDeviceModel: `emb_feature_indices`의 임베딩 피쳐 인덱스 길이가 emb_x 임베딩 피쳐 길이보다 큽니다.")

        # - 1.2. None 피쳐 기본값 할당 및 임베딩 피쳐 펼치기
        non_emb_x = x[: len(x) - len(self.default_emb_feature_values)]
        non_emb_x = [f if f is not None else self.default_model_feature_values[i] for i, f in enumerate(non_emb_x)]
        emb_x = [
            [
                emb_x[i][j] if emb_x[i][j] is not None else self.default_emb_feature_values[i][j]
                for j in self.emb_feature_indices[i]
            ]
            for i in range(len(emb_x))
        ]
        emb_x = reduce(lambda a, b: a + b, emb_x, [])

        return self._convert(non_emb_x) + emb_x

    def _ml_predict(self, x: List[Any]) -> List[float]:
        return self.models[0].predict(np.array([x]))[0].tolist()

    def _postprocess(self, y: List[float], user_mdl_cd: str, user_age: int) -> List[Dict[str, Any]]:
        user_device_info = self.device_meta.get(user_mdl_cd, {})

        user_grp_id = user_device_info.get("product_grp_id", "")
        user_grp_nm = user_device_info.get("product_grp_nm", "")
        user_mfact_nm = user_device_info.get("mfact_nm", "")

        results = []
        for label_index, label in self.index_labels.items():
            try:
                prod_list = self.product_classes[label]
                for prod in prod_list:
                    prod.update({"score": y[label_index]})
                    results.append(prod)
            except Exception:
                raise MLSModelError("`label`이 product_classes에 존재하지 않습니다.")

        items = []
        for result in results:
            rec_grp_nm = re.sub("5G|Re", "", result["product_grp_nm"])
            high_rank_eqp_nm_list = [(rec_grp_nm + " " + word).replace("  ", " ") for word in self.high_rank_word]

            # 본인 단말기 제외
            if result["product_grp_id"] == user_grp_id:
                score = 0
            # 애플 고객에게는 아이폰만 추천
            elif user_mfact_nm == "Apple" and result["mfact_nm"] == "Apple":
                score = 50 * result["device_weight"]
            # 현재 단말기가 울트라, pro 등 상위 단말기일 경우 제외
            elif user_grp_nm in high_rank_eqp_nm_list:
                score = 0
            else:
                score = result["score"] * result["device_weight"]

            # 컨텍스트 빌드
            context_features = self.products.get(result["product_grp_id"], {}).get("context_features", [])
            context_id = self.default_context_value
            if context_features:
                for context_feature in context_features:
                    if re.match("^context_age[0-9]{2}", context_feature):
                        try:
                            age_bin = int(context_feature.split("_")[1][3:5])
                            if int(user_age) // 10 * 10 == age_bin:
                                context_id = context_feature
                                break
                        except Exception:
                            break
                    else:
                        context_id = context_feature
                        break

            items.append(
                {
                    "id": result["product_grp_id"],
                    "name": result["product_grp_nm"],
                    "score": score,
                    "props": {"context_id": context_id},
                    "type": "device",
                }
            )

        items = sorted(items, key=lambda x: x["score"], reverse=True)[: min(self.num_pick, len(items))]
        return [{k: v for k, v in d.items() if k != "score"} for d in items]

    def _convert(self, features):
        feature_names = self.features[2 : len(self.features) - len(self.default_emb_feature_values)]
        converted_features = []
        for i, f in enumerate(features):
            feature_name = feature_names[i]
            formula = self.conversion_formulas.get(feature_name)
            if formula:
                if "map" in formula:
                    converted_features.append(formula["map"].get(f, formula["default"]))
                elif "conditions" in formula:
                    conditions = formula["conditions"]
                    conditions_checked = True
                    for condition in conditions:
                        if (
                            (condition[0] == "==" and f != condition[1])
                            or (condition[0] == "!=" and f == condition[1])
                            or (condition[0] == "<" and f >= condition[1])
                            or (condition[0] == "<=" and f > condition[1])
                            or (condition[0] == ">" and f <= condition[1])
                            or (condition[0] == ">=" and f < condition[1])
                        ):
                            conditions_checked = False
                            break
                    converted_features.append(f if conditions_checked else formula["default"])
            else:
                try:
                    converted_feature = float({"Y": 1, "N": 0}.get(f, f))
                except Exception:
                    converted_feature = {"Y": 1, "N": 0}.get(f, f)

                converted_features.append(converted_feature)
        return converted_features



import pytest

import numpy as np

import lightgbm

from sktmls.models.contrib import LightGBMRandomContextModel


class TestLightGBMRandomContextModels:
    @pytest.fixture(scope="class")
    def param(self):
        return {
            "model": LightGBMRandomContextModel(
                model=lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
                model_name="test_model",
                model_version="test_version",
                model_features=["feature1", "feature2", "feature3", "feature4"],
                default_feature_values=["3", 1, "A", 100],
                context_features=["feature5", "feature2"],
                context_values=["context_1", "context_2"],
                default_context_value="default_context",
                cutoff=0.3,
                item_id="id",
                item_name="name",
                item_type="type",
            )
        }

    def test_000_lightgbm_random_context_return(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.2]))
        assert param.get("model").predict([0, 0, 0, 0, "Y", "Y"]) == {"items": []}

    def test_001_lightgbm_random_context_return(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.8]))
        assert param.get("model").predict([999, 999, 999, 999, "N", "N"]) == {
            "items": [
                {
                    "id": "id",
                    "name": "name",
                    "type": "type",
                    "props": {"score": 0.8, "context_id": "default_context"},
                }
            ]
        }

    def test_002_lightgbm_random_context_return(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.8]))

        results = param.get("model").predict([999, 999, 999, 999, "Y", "N"])

        assert "items" in results
        items = results["items"][0]

        assert "props" in items
        props = items.pop("props")

        assert items == {
            "id": "id",
            "name": "name",
            "type": "type",
        }

        assert props.get("score") == 0.8
        assert props.get("context_id") in ["context_1", "context_2"]

    def test_003_lightgbm_random_context_return(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.8]))

        results = param.get("model").predict([999, 999, 999, 999, "N", "Y"])

        assert "items" in results
        items = results["items"][0]

        assert "props" in items
        props = items.pop("props")

        assert items == {
            "id": "id",
            "name": "name",
            "type": "type",
        }

        assert props.get("score") == 0.8
        assert props.get("context_id") in ["context_1", "context_2"]

    def test_004_lightgbm_random_context_return(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.8]))

        results = param.get("model").predict([999, 999, 999, 999, "Y", "N"])

        assert "items" in results
        items = results["items"][0]

        assert "props" in items
        props = items.pop("props")

        assert items == {
            "id": "id",
            "name": "name",
            "type": "type",
        }

        assert props.get("score") == 0.8
        assert props.get("context_id") in ["context_1", "context_2"]
